# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** CIRES, Regents of the University of Colorado
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
import numpy as np
import xarray as xr
import scipy
import xrft


def zonal_wave_coeffs(dat, dimvar='longitude', *, waves=None, fftpkg='scipy'):
    r"""Calculate the Fourier coefficients of waves in the zonal direction.
    This is a primarily a driver function that shifts the data depending
    on the specified fftpkg.
    Parameters
    ----------
    dat : `xarray.DataArray`
        data containing a dimension over which to compute coefficients that spans 
        all 360 degrees
    dimvar: Name of the dimension to compute the wave coefficients.  Longitude is the
        default if it's not specified
    waves : array-like, optional
        The zonal wavenumbers to maintain in the output. Defaults to None for all.
    fftpkg : string, optional
        String that specifies how to perform the FFT on the data. Options are
        scipy or xrft. Specifying scipy uses some operations that are memory-eager
        and leverages scipy.fft.rfft. Specifying xrft should leverage the benefits
        of xarray/dask for large datasets by using xrft.fft. Defaults to scipy.
    Returns
    -------
    `xarray.DataArray`
        Output of the rFFT along the longitude dimension, for specified waves only.
    """

    if fftpkg not in ['scipy', 'xrft']:
        msg = 'fftpkg keyword arg must be one of scipy or xarray'
        raise ValueError(msg)

    funcs = {
        'scipy': _zonal_wave_coeffs_scipy,
        'xrft': _zonal_wave_coeffs_xrft
    }

    nlons = dat[dimvar].size

    fc = funcs[fftpkg](dat, dimvar)

    fc.attrs['nlons'] = nlons
    fc.attrs['lon0'] = dat[dimvar].values[0]
    if (waves is not None):
        fc = fc.sel(lon_wavenum=waves)

    return fc


def _zonal_wave_coeffs_scipy(dat, dimvar):
    r"""Calculate the Fourier coefficients of waves in the zonal direction.
    Uses scipy.fft.rfft to perform the calculation.
    Parameters
    ----------
    dat : `xarray.DataArray`
        data containing a dimension over which to compute coefficients that spans
        all 360 degrees
    dimvar: Name of the dimension to compute the wave coefficients.
    Returns
    -------
    `xarray.DataArray`
        Output of the rFFT along the longitude dimension.
    """
    nlons = dat[dimvar].size
    lon_ax = dat.get_axis_num(dimvar)

    new_dims = list(dat.dims)
    new_dims[lon_ax] = 'lon_wavenum'

    new_coords = dict(dat.coords)
    new_coords.pop(dimvar)
    new_coords['lon_wavenum'] = np.arange(0, nlons//2 + 1)

    fc = scipy.fft.rfft(dat.values, axis=lon_ax)
    fc = xr.DataArray(fc, coords=new_coords, dims=new_dims)

    return fc


def _zonal_wave_coeffs_xrft(dat, dimvar):
    r"""Calculate the Fourier coefficients of waves in the zonal direction.
    Uses xrft.fft to perform the calculation.
    Parameters
    ----------
    dat : `xarray.DataArray`
        data containing a dimension over which to compute coefficients that spans
        all 360 degrees
    dimvar: Name of the dimension to compute the wave coefficients.
    Returns
    -------
    `xarray.DataArray`
        Output of the rFFT along the longitude dimension.
    """

    fc = xrft.fft(dat, dim=dimvar, real_dim=dimvar,
                  true_phase=False, true_amplitude=False)

    fc = fc.rename({'freq_longitude': 'lon_wavenum'})
    fc = fc.assign_coords({'lon_wavenum': np.arange(fc.lon_wavenum.size)})

    return fc


def zonal_wave_ampl_phase(dat, dimvar='longitude', waves=None, phase_deg=False, fftpkg='scipy'):
    r"""Calculates the amplitudes and relative phases of waves in the zonal direction.
    Parameters
    ----------
    dat : `xarray.DataArray`
        data containing a dimension to compute wave amplitudes and phases that spans 
        all 360 degrees
    dimvar: Name of the dimension to compute the amplitude and phase.  Longitude is the
        default if it's not specified
    waves : array-like, optional
        The zonal wavenumbers to maintain in the output. Defaults to None for all.
    phase_deg : boolean, optional
        Whether to return the relative phases in radians or degrees.
    fftpkg : string, optional
        String that specifies how to perform the FFT on the data. Options are
        scipy or xrft. Specifying scipy uses some operations that are memory-eager
        and leverages scipy.fft.rfft. Specifying xrft should leverage the benefits
        of xarray/dask for large datasets by using xrft.fft. Defaults to scipy.
    Returns
    -------
    Tuple of two `xarray.DataArray`
        Tuple contains (amplitudes, phases)
    See Also
    --------
    zonal_wave_coeffs
    """

    fc = zonal_wave_coeffs(dat, dimvar, waves=waves, fftpkg=fftpkg)

    # where the longitudinal wavenumber is 0, `where' will
    # mask to nan, so np.isfinite will return False in those
    # spots and true everywhere else. Thus, add 1 to get
    # the multiplying mask that keeps in mind the "zeroth"
    # mode (which would be the zonal mean, if kept)
    #
    # this is necessary because of the symmetric spectrum,
    # so all other wavenumbers except the 0th need to
    # be multipled by 2 to get the right amplitude
    mult_mask = np.isfinite(fc.where(fc.lon_wavenum != 0)) + 1

    ampl = mult_mask*np.abs(fc) / fc.nlons
    phas = np.angle(fc, deg=phase_deg)

    return (ampl.astype(dat.dtype), phas.astype(dat.dtype))


def zonal_wave_contributions(dat, dimvar='longitude', waves=None, fftpkg='scipy'):
    r"""Computes contributions of waves with zonal wavenumber k to the input field.
    Parameters
    ----------
    dat : `xarray.DataArray`
        data containing a dimension to compute wave contributions that spans all 
        360 degrees
    dimvar: Name of the dimension to compute the contributions of waves.  Longitude 
        is the default if it's not specified
    waves : array-like, optional
        The zonal wavenumbers to maintain in the output. Defaults to None for all.
    fftpkg : string, optional
        String that specifies how to perform the FFT on the data. Options are
        scipy or xrft. Specifying scipy uses some operations that are memory-eager
        and leverages scipy.fft.rfft. Specifying xrft should leverage the benefits
        of xarray/dask for large datasets by using xrft.fft. Defaults to scipy.
    Returns
    -------
    `xarray.DataArray`
    See Also
    --------
    zonal_wave_coeffs
    """
    fc = zonal_wave_coeffs(dat, dimvar, waves=waves, fftpkg=fftpkg)

    if (waves is None):
        waves = fc.lon_wavenum.values

    recons = []
    if (fftpkg == 'scipy'):
        new_dims = list(dat.dims)
        new_dims += ['lon_wavenum']
        new_coords = dict(dat.coords)
        new_coords['lon_wavenum'] = waves

        for k in waves:
            mask = np.isnan(fc.where(fc.lon_wavenum != k))

            kcont = scipy.fft.irfft((fc*mask).values, axis=fc.get_axis_num('lon_wavenum'))
            recons.append(kcont[..., np.newaxis])

        recons = np.concatenate(recons, axis=-1)
        recons = xr.DataArray(recons, dims=new_dims, coords=new_coords)

    elif (fftpkg == 'xarray'):
        fc = fc.rename({'lon_wavenum': 'freq_longitude'})

        for k in waves:
            mask = np.isnan(fc.where(fc.lon_wavenum != k))

            kcont = xrft.ifft((fc*mask).values, dim='lon_wavenum', real_dim='lon_wavenum')
            recons.append(kcont)

        recons = xr.concat(recons, dim='lon_wavenum')
        recons = recons.assign_coords({'lon_wavenum': waves, dimvar: dat[dimvar]})

    return recons.astype(dat.dtype)


def zonal_wave_covariance(dat1, dat2, dimvar1='longitude', dimvar2='longitude', waves=None, fftpkg='scipy'):
    r"""Calculates the covariance of two fields partititioned into zonal wavenumbers.
    Parameters
    ----------
    dat1 : `xarray.DataArray`
        field containing a dimension over which to compute wave covariance that 
        spans all 360 degrees.  Should have the same shape as dat2.
    dat2 : `xarray.DataArray`
        another field also containing a dimension over which to compute wave covariance 
        that spans all 360 degrees. Should have the same shape as dat1.
    dimvar1: Name of the dimension to compute the wave covariance for dat1.  Longitude 
        is the default if it's not specified
    dimvar2: Name of the dimension to compute the wave ccovariance for dat2.  Longitude 
        is the default if it's not specified
    waves : array-like, optional
        The zonal wavenumbers to maintain in the output. Defaults to None for all.
    fftpkg : string, optional
        String that specifies how to perform the FFT on the data. Options are
        scipy or xrft. Specifying scipy uses some operations that are memory-eager
        and leverages scipy.fft.rfft. Specifying xrft should leverage the benefits
        of xarray/dask for large datasets by using xrft.fft. Defaults to scipy.
    Returns
    -------
    `xarray.DataArray`
    See Also
    --------
    zonal_wave_coeffs
    TO DO
    -----
    * Check for consistency between dat1 and dat2 and throw errors
    """

    nlons = dat1[dimvar].size

    fc1 = zonal_wave_coeffs(dat1, dimvar1, waves=waves, fftpkg=fftpkg)
    fc2 = zonal_wave_coeffs(dat2, dimvar2, waves=waves, fftpkg=fftpkg)

    mult_mask = np.isfinite(fc1.where(fc1.lon_wavenum != 0)) + 1
    cov = mult_mask*np.real(fc1 * fc2.conj())/(nlons**2)

    return cov



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: ctc_statistics.py
"""
import warnings
from typing import Union
import math
import re
import numpy as np
import pandas as pd
from scipy.special import lambertw
from metcalcpy.util.utils import round_half_up, column_data_by_name, \
    sum_column_data_by_name, PRECISION

__author__ = 'Tatiana Burek'
__version__ = '0.1.0'


def calculate_baser(input_data, columns_names):
    """Performs calculation of BASER - Base rate, aka Observed relative frequency

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated BASER as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')

    try:
        result = (sum(input_data[:, np.where(columns_names == 'fy_oy')[0][0]])
                  + sum(input_data[:, np.where(columns_names == 'fn_oy')[0][0]])) \
                 / sum(input_data[:, np.where(columns_names == 'total')[0][0]])
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_acc(input_data, columns_names):
    """Performs calculation of ACC - Accuracy

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated ACC as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        result = (sum_column_data_by_name(input_data, columns_names, 'fy_oy')
                  + sum_column_data_by_name(input_data, columns_names, 'fn_on')) \
                 / sum_column_data_by_name(input_data, columns_names, 'total')
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_fbias(input_data, columns_names):
    """Performs calculation of FBIAS - Bias, aka Frequency bias

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated FBIAS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy') \
             + sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        if oy == 0:
            return None
        oyn = sum_column_data_by_name(input_data, columns_names, 'fy_oy') \
              + sum_column_data_by_name(input_data, columns_names, 'fy_on')
        result = oyn / oy
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_fmean(input_data, columns_names):
    """Performs calculation of FMEAN - Forecast mean

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated FMEAN as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = sum_column_data_by_name(input_data, columns_names, 'total')
        if total == 0:
            return None
        oyn = sum_column_data_by_name(input_data, columns_names, 'fy_oy') \
              + sum_column_data_by_name(input_data, columns_names, 'fy_on')
        result = oyn / total
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_pody(input_data, columns_names):
    """Performs calculation of PODY - Probability of Detecting Yes

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated PODY as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')

    try:
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        oy = fy_oy + fn_oy
        result = fy_oy / oy
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_pofd(input_data, columns_names):
    """Performs calculation of POFD - Probability of false detection

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated POFD as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        oy = fy_on + sum_column_data_by_name(input_data, columns_names, 'fn_on')
        result = fy_on / oy
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ctc_roc(data, ascending):
    """ Creates a data frame to hold the aggregated contingency table and ROC data
            Args:
                :param data: pandas data frame with ctc data and column names:
                    - fcst_thresh
                    - fy_oy
                    - fy_on
                    - fn_oy
                    - fn_on
                    - fcst_valid_beg
                    - fcst_lead
                :param ascending: order in which to sort the input data by fcst_thresh. Default is
                                  True, set to False to sort by descending order.

            Returns:
                pandas data frame with ROC data and columns:
                - thresh
                - pody
                - pofd
    """
    # create a data frame to hold the aggregated contingency table and ROC data
    sorted_data = sort_by_thresh(data, ascending=ascending)
    list_thresh = np.sort(np.unique(sorted_data['fcst_thresh'].to_numpy()))

    # If descending order was requested for sorting the input dataframe,
    # reverse the order of the list_thresh to
    # maintain results in descending order.
    if not ascending:
        list_thresh = list_thresh[::-1]

    df_roc = pd.DataFrame(
        {'thresh': list_thresh, 'pody': None, 'pofd': None})

    index = 0
    for thresh in list_thresh:
        # create a subset of the sorted_data that contains only the rows of the unique
        # threshold values
        subset_data = sorted_data[sorted_data['fcst_thresh'] == thresh]
        data_np = subset_data.to_numpy()
        columns = subset_data.columns.values
        pody = calculate_pody(data_np, columns)
        pofd = calculate_pofd(data_np, columns)
        df_roc.loc[index] = [thresh, pody, pofd]
        index += 1

    return df_roc


def calculate_podn(input_data, columns_names):
    """Performs calculation of PODN - Probability of Detecting No

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated PODN as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
        oy = sum_column_data_by_name(input_data, columns_names, 'fy_on') + fn_on
        result = fn_on / oy
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_far(input_data, columns_names):
    """Performs calculation of FAR - false alarms

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated FAR as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy') + fy_on
        result = fy_on / oy
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_csi(input_data, columns_names):
    """Performs calculation of CSI - Critical success index, aka Threat score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated CSI as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        oy = fy_oy \
             + sum_column_data_by_name(input_data, columns_names, 'fy_on') \
             + sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        result = fy_oy / oy
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_gss(input_data, columns_names):
    """Performs calculation of GSS = Gilbert skill score, aka Equitable threat score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated GSS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = sum_column_data_by_name(input_data, columns_names, 'total')
        if total == 0:
            return None
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        dbl_c = ((fy_oy + fy_on) / total) * (fy_oy + fn_oy)
        gss = ((fy_oy - dbl_c) / (fy_oy + fy_on + fn_oy - dbl_c))
        gss = round_half_up(gss, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        gss = None
    warnings.filterwarnings('ignore')
    return gss


def calculate_hk(input_data, columns_names):
    """Performs calculation of HK - Hanssen Kuipers Discriminant

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated HK as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        pody = calculate_pody(input_data, columns_names)
        pofd = calculate_pofd(input_data, columns_names)
        if pody is None or pofd is None:
            result = None
        else:
            result = pody - pofd
            result = round_half_up(result, PRECISION)
    except (TypeError, Warning):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_hss(input_data, columns_names):
    """Performs calculation of HSS - Heidke skill score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated HSS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        total = sum_column_data_by_name(input_data, columns_names, 'total')
        if total == 0:
            return None
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
        dbl_c = ((fy_oy + fy_on) / total) * (fy_oy + fn_oy) + ((fn_oy + fn_on) / total) * (fy_on + fn_on)
        hss = ((fy_oy + fn_on - dbl_c)/ (total - dbl_c))
        hss = round_half_up(hss, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        hss = None
    warnings.filterwarnings('ignore')
    return hss


def calculate_odds(input_data, columns_names):
    """Performs calculation of ODDS - Odds Ratio

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated ODDS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        pody = calculate_pody(input_data, columns_names)
        pofd = calculate_pofd(input_data, columns_names)
        if pody is None or pofd is None:
            result = None
        else:

            result = (pody * (1 - pofd)) / (pofd * (1 - pody))
            result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')

    return result


def calculate_lodds(input_data, columns_names):
    """Performs calculation of LODDS - Log Odds Ratio

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated LODDS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
        if fy_oy is None or fy_on is None or fn_oy is None or fn_on is None:
            return None
        v = np.log(fy_oy) + np.log(fn_on) - np.log(fy_on) - np.log(fn_oy)
        v = round_half_up(v, PRECISION)
    except (TypeError, Warning):
        v = None
    warnings.filterwarnings('ignore')
    return v


def calculate_bagss(input_data, columns_names):
    """Performs calculation of BAGSS - Bias-Corrected Gilbert Skill Score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated BAGSS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        total = sum_column_data_by_name(input_data, columns_names, 'total')
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        if fy_oy is None or fn_oy is None or fy_on is None or total is None:
            return None
        if fy_oy == 0 or fn_oy == 0 or total == 0:
            return None
        dbl_o = fy_oy + fn_oy
        dbl_lf = np.log(dbl_o / fn_oy)
        lambert = lambertw(dbl_o / fy_on * dbl_lf).real
        dbl_ha = dbl_o - (fy_on / dbl_lf) * lambert
        result = (dbl_ha - (dbl_o ** 2 / total)) / (2 * dbl_o - dbl_ha - (dbl_o ** 2 / total))
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_eclv(input_data, columns_names):
    """Performs calculation of ECLV - Economic Cost/Loss  Value
        Implements R version that returns an array instead of the single value
        IT WILL NOT WORK - NEED TO CONSULT WITH STATISTICIAN
        Build list of X-axis points between 0 and 1

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated BAGSS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        cl_step = 0.05
        cl_pts = np.arange(start=cl_step, stop=1, step=cl_step)
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
        eclv = calculate_economic_value(np.array([fy_oy, fy_on, fn_oy, fn_on]), cl_pts)
        common_cases_ind = pd.Series(eclv['cl']).isin(cl_pts)
        v = eclv['V'][common_cases_ind]
        v = round_half_up(v, PRECISION)
    except (TypeError, Warning):
        v = None
    warnings.filterwarnings('ignore')
    return v


def calculate_economic_value(values, cost_lost_ratio=np.arange(start=0.05, stop=0.95, step=0.05),
                             add_base_rate: bool = False) -> Union[dict, None]:
    """Calculates the economic value of a forecast based on a cost/loss ratio.
       Similar to R script function 'value' from  the 'verification' package

        Args:
            values: An array vector of a contingency table summary of values in the form
                c(n11, n01, n10, n00) where in nab a = obs, b = forecast.
            cost_lost_ratio:  Cost loss ratio. The relative value of being unprepared
                and taking a loss to that of un-necessarily preparing. For example,
                cl = 0.1 indicates it would cost $ 1 to prevent a $10 loss.
                This defaults to the sequence 0.05 to 0.95 by 0.05.
            add_base_rate: add Base rate point to cl or not

        Returns:
            If assigned to an object, the following values are reported in the dictionary :
                vmax - Maximum value
                V    - Vector of values for each cl value
                F    - Conditional false alarm rate.
                H    - Conditional hit rate
                cl   - Vector of cost loss ratios.
                s    - Base rate
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        if len(values) == 4:
            n = sum(values)
            f = values[1] / (values[1] + values[3])
            h = values[0] / (values[0] + values[2])
            s = (values[0] + values[2]) / n

            if add_base_rate is True:
                cl_local = np.append(cost_lost_ratio, s)
            else:
                cl_local = np.copy(cost_lost_ratio)
            cl_local.sort()

            v_1 = (1 - f) - s / (1 - s) * (1 - cl_local) / cl_local * (1 - h)
            v_2 = h - (1 - s) / s * cl_local / (1 - cl_local) * f
            v = np.zeros(len(cl_local))

            indexes = cl_local < s
            v[indexes] = v_1[indexes]
            indexes = cl_local >= s
            v[indexes] = v_2[indexes]

            v_max = h - f
            result = {'vmax': round_half_up(v_max, PRECISION),
                      'V': v,
                      'F': round_half_up(f, PRECISION),
                      'H': round_half_up(h, PRECISION),
                      'cl': cl_local,
                      's': round_half_up(s, PRECISION),
                      'n': n}
        else:
            result = None
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_ctc_total(input_data, columns_names):
    """Calculates the Total number of matched pairs for Contingency Table Counts

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Total number of matched pairs as float
            or None if some of the data values are missing or invalid
    """
    total = sum_column_data_by_name(input_data, columns_names, 'total')
    return round_half_up(total, PRECISION)


def calculate_cts_total(input_data, columns_names):
    """Calculates the Total number of matched pairs for Contingency Table Statistics

            Args:
                input_data: 2-dimensional numpy array with data for the calculation
                    1st dimension - the row of data frame
                    2nd dimension - the column of data frame
                columns_names: names of the columns for the 2nd dimension as Numpy array

            Returns:
                calculated Total number of matched pairs as float
                or None if some of the data values are missing or invalid
    """
    total = sum_column_data_by_name(input_data, columns_names, 'total')
    return round_half_up(total, PRECISION)


def calculate_ctc_fn_on(input_data, columns_names):
    """Calculates the Number of forecast no and observation no for Contingency Table Statistics

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Number of forecast no and observation no as float
            or None if some of the data values are missing or invalid
    """
    fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
    return round_half_up(fn_on, PRECISION)


def calculate_ctc_fn_oy(input_data, columns_names):
    """Calculates the Number of forecast no and observation yes for Contingency Table Statistics

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Number of forecast no and observation yes as float
            or None if some of the data values are missing or invalid
    """
    fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
    return round_half_up(fn_oy, PRECISION)


def calculate_ctc_fy_on(input_data, columns_names):
    """Calculates the Number of forecast yes and observation no for Contingency Table Statistics

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Number of forecast yes and observation no as float
            or None if some of the data values are missing or invalid
    """
    fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
    return round_half_up(fy_on, PRECISION)


def calculate_ctc_fy_oy(input_data, columns_names):
    """Calculates the Number of forecast yes and observation yes for Contingency Table Statistics

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated Number of forecast yes and observation yes as float
            or None if some of the data values are missing or invalid
    """
    fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
    return round_half_up(fy_oy, PRECISION)


def calculate_ctc_oy(input_data, columns_names):
    """Calculates the Total Number of forecast yes and observation yes plus
        Number of forecast no and observation yes for Contingency Table Statistics

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated OY as float
            or None if some of the data values are missing or invalid
    """
    fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
    fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
    return round_half_up(fy_oy + fn_oy, PRECISION)


def calculate_ctc_on(input_data, columns_names):
    """Calculates the Total Number of forecast yes and observation no plus
        Number of forecast no and observation no for Contingency Table Statistics

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated ON as float
            or None if some of the data values are missing or invalid
    """
    fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
    fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
    return round_half_up(fy_on + fn_on, PRECISION)


def calculate_ctc_fy(input_data, columns_names):
    """Calculates the Total Number of forecast yes and observation no plus
        Number of forecast yes and observation yes for Contingency Table Statistics

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated FY as float
            or None if some of the data values are missing or invalid
    """
    fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
    fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
    return round_half_up(fy_on + fy_oy, PRECISION)


def calculate_ctc_fn(input_data, columns_names):
    """Calculates the Total Number of forecast no and observation no plus
        Number of forecast no and observation yes for Contingency Table Statistics

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated FN as float
            or None if some of the data values are missing or invalid
    """
    fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
    fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
    return round_half_up(fn_on + fn_oy, PRECISION)


def pod_yes(input_data, columns_names):
    warnings.filterwarnings('error')
    try:
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        num = fy_oy
        den = fy_oy + sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        result = num / den
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def pod_no(input_data, columns_names):
    warnings.filterwarnings('error')
    try:
        fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
        num = fn_on
        den = fn_on + sum_column_data_by_name(input_data, columns_names, 'fy_on')
        result = num / den
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_odds1(input_data, columns_names):
    """Performs calculation of ODDS - Odds Ratio

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated ODDS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        py = pod_yes(input_data, columns_names)
        pn = calculate_pofd(input_data, columns_names)

        num = py / (1 - py)
        den = pn / (1 - pn)
        result = num / den
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_orss(input_data, columns_names):
    """Performs calculation of ORSS - Odds Ratio Skill Score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated ORSS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')

        num = fy_oy * fn_on - fy_on * fn_oy
        den = fy_oy * fn_on + fy_on * fn_oy
        result = num / den
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_sedi(input_data, columns_names):
    """Performs calculation of SEDI - Symmetric Extremal Depenency Index

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated SEDI as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')

        f = fy_on / (fy_on + fn_on)
        h = pod_yes(input_data, columns_names)

        num = math.log(f) - math.log(h) - math.log(1 - f) + math.log(1 - h)
        den = math.log(f) + math.log(h) + math.log(1 - f) + math.log(1 - h)
        result = num / den
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_seds(input_data, columns_names):
    """Performs calculation of SEDS - Symmetric Extreme Dependency Score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated SEDS as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        total = sum_column_data_by_name(input_data, columns_names, 'total')

        num = math.log((fy_oy + fy_on) / total) + math.log((fy_oy + fn_oy) / total)

        den = math.log(fy_oy / total)
        result = num / den - 1.0
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_edi(input_data, columns_names):
    """Performs calculation of EDI - Extreme Dependency Index

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated EDI as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fn_on = sum_column_data_by_name(input_data, columns_names, 'fn_on')
        fy_on = sum_column_data_by_name(input_data, columns_names, 'fy_on')
        total = sum_column_data_by_name(input_data, columns_names, 'total')
        f = fy_on / (fy_on + fn_on)
        h = pod_yes(input_data, columns_names)

        num = math.log(f) - math.log(h)
        den = math.log(f) + math.log(h)
        result = num / den
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def calculate_eds(input_data, columns_names):
    """Performs calculation of EDS - Extreme Dependency Score

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated EDs as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')
    try:
        fy_oy = sum_column_data_by_name(input_data, columns_names, 'fy_oy')
        fn_oy = sum_column_data_by_name(input_data, columns_names, 'fn_oy')
        total = sum_column_data_by_name(input_data, columns_names, 'total')

        num = math.log((fy_oy + fn_oy) / total)
        den = math.log(fy_oy / total)

        result = 2.0 * num / den - 1.0
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result


def sort_by_thresh(input_dataframe: pd.DataFrame, sort_column_name: str = 'fcst_thresh',
                   ascending: bool = True) -> pd.DataFrame:
    """
        Sorts the input pandas dataframe by threshold values in the specified column that have
        format "operator value", ie >=1.  This is done by first separating
        the fcst_thresh column into a threshold operator (<,<=, ==, >=, >)
        and thresh value column.  Assign a weight to each operator (1 for <,
        2 for <=, 3 for ==  and no operator, 4 for >=, and 5 for >).  Finally,
        sort the input dataframe by these two new columns resulting in a new
        dataframe sorted by fcst_thresh in ascending order (default) or descending
        order.

        Args:

            :param input_dataframe: A pandas dataframe representing data that is to be
                                   sorted according to the specifiec column name.


            :param sort_column_name:  The column to base the sorting.  The default is 'fcst_thresh'
                                     (which is a colunn in CTC output)

            :param ascending: A boolean value, by default is set to True to sort by
                              ascending value.  Set to False to sort in descending order.

        Returns:
            sorted_df:  A pandas dataframe that is sorted based on the specified column's values and
                        in the specified order (ascending or descending-default is ascending).


    """

    operators = []
    values = []
    text_strings = []
    requested_thresh = input_dataframe[sort_column_name]

    for thresh in requested_thresh:
        # for thresholds that are comprised of an operator and value, ie >=3,
        # separate the fcst_thresh into two parts: the operator (ie <, <=, ==, >=, >)
        # and the numerical value of the threshold (which can be a negative value)
        match = re.match(r'(\<|\<=|\==|\>=|\>)*((-)*([0-9])(.)*)', thresh)
        match_text = re.match(r'(\<|\<=|\==|\>=|\>)*((.)*)', thresh)

        if match:
            operators.append(match.group(1))
            value = float(match.group(2))
            values.append(value)
        elif match_text:
            operators.append(match_text.group(1))
            text = match_text.group(2)
            text_strings.append(text)

    # apply a numerical weighting to each operator: 1 for <, 2 for <=, etc.
    wt_maps = {'<': 1, '<=': 2, '==': 3, '>=': 4, '>': 5}
    operator_wts = []

    for operator in operators:
        # if no operator is found, assign the same
        # weight as used for the == operator (i.e. assume
        # that if a bare number is observed, assume that
        # a == is implied)
        if operator is None:
            # no operator, assume ==
            operator_wts.append(wt_maps['=='])
        else:
            operator_wts.append(wt_maps[operator])

    # Columns to use in pandas dataframe sorting
    sort_by_cols = ['thresh_values', 'op_wts']
    input_dataframe['op_wts'] = operator_wts

    # assign new column based on the format of the threshold values
    if match:
        input_dataframe['thresh_values'] = values
    elif match_text:
        input_dataframe['thresh_values'] = text_strings
    else:
        # if the threshold values don't conform to what is expected, then
        # use the requested threshold column name for sorting.
        sort_by_cols = [sort_column_name]

    # sort with ignore_index=True because we don't need to keep the original index values. We
    # want the rows to be newly indexed to reflect the reordering. Use inplace=False because
    # we don't want to modify the input dataframe's order, we want a new dataframe.
    sorted_dataframe = input_dataframe.sort_values(by=sort_by_cols, inplace=False,
                                                   ascending=ascending, ignore_index=True)
    return sorted_dataframe



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Read in OMI or RMM indices and plot phase diagram for specified dates. OMI values
can be obtained from https://psl.noaa.gov/mjo/, RMM values can be obtained from 
http://www.bom.gov.au/climate/mjo/graphics/rmm.74toRealtime.txt
"""

import numpy as np
import pandas as pd
import datetime

from plot_mjo_indices import phase_diagram

# which index are we plotting
indexname = 'RMM'  # 'RMM' or 'OMI'

# set dates to read and plot
datestrt = 20120101
datelast = 20120331

# read data from text file
if indexname=='OMI':
    data = pd.read_csv('./data/omi.1x.txt', header=None, delim_whitespace=True, names=['yyyy','mm','dd','hh','pc1','pc2','amp'])
elif indexname=='RMM':
    data = pd.read_csv('./data/rmm.1x.txt',  header=None, delim_whitespace=True, names=['yyyy','mm','dd', 'pc1','pc2','phase','amp','source'])

DATES = data.yyyy.values*10000 + data.mm.values*100 + data.dd.values
MONTHS = data.mm.values
DAYS = data.dd.values
#print(dates)

istrt = np.where(DATES==datestrt)[0][0]
ilast = np.where(DATES==datelast)[0][0]
print(DATES[istrt], DATES[ilast])
#print(istrt, ilast)

# subset data to only the dates we want to plot
dates = DATES[istrt:ilast+1]
months = MONTHS[istrt:ilast+1]
days = DAYS[istrt:ilast+1]
print(dates.min(), dates.max())
PC1 = data.pc1.values[istrt:ilast+1]
PC2 = data.pc2.values[istrt:ilast+1]

# plot the phase diagram
phase_diagram(indexname,PC1,PC2,dates,months,days,indexname+'_phase','png')


# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Compute RMM index from input U850, U200 and OLR data. Data is averaged from 20S-20N
"""

import numpy as np
import xarray as xr
import datetime
import pandas as pd

import compute_mjo_indices as cmi
import plot_mjo_indices as pmi

# set dates to read
datestrt = '2000-01-01'
datelast = '2002-12-31'

spd = 1 # number of obs per day
time = np.arange(datestrt,datelast, dtype='datetime64[D]')
ntim = len(time)

#######################################
# read RMM EOFs from file and plot
EOF1, EOF2 = cmi.read_rmm_eofs('./data/')
pmi.plot_rmm_eofs(EOF1, EOF2, 'RMM_EOFs','png')

#######################################
# read data from file
ds = xr.open_dataset('/data/mgehne/OLR/olr.1x.7920.anom7901.nc')
olr = ds['olr'].sel(lat=slice(-15,15),time=slice(datestrt,datelast))
lon = ds['lon']
olr = olr.mean('lat')
print(olr.min(), olr.max())

ds = xr.open_dataset('/data/mgehne/ERAI/uwnd.erai.an.2p5.850.daily.anom7901.nc')
u850 = ds['uwnd'].sel(lat=slice(-15,15),time=slice(datestrt,datelast))
u850 = u850.mean('lat')
print(u850.min(), u850.max())

ds = xr.open_dataset('/data/mgehne/ERAI/uwnd.erai.an.2p5.200.daily.anom7901.nc')
u200 = ds['uwnd'].sel(lat=slice(-15,15),time=slice(datestrt,datelast))
u200 = u200.mean('lat')
print(u200.min(), u200.max())

########################################
# project data onto EOFs
PC1, PC2 = cmi.rmm(olr[0:ntim,:], u850[0:ntim,:], u200[0:ntim,:], time, spd, './data/')

print(PC1.min(), PC1.max())


########################################
# plot phase diagram
datestrt = '2002-01-01'
datelast = '2002-12-31'

time = np.arange(datestrt,datelast, dtype='datetime64[D]')
ntim = len(time)
PC1 = PC1.sel(time=slice(datestrt,datelast))
PC2 = PC2.sel(time=slice(datestrt,datelast))
PC1 = PC1[0:ntim]
PC2 = PC2[0:ntim]

months = []
days = []
for idx, val in enumerate(time):
    date = pd.to_datetime(val).timetuple()
    month = date.tm_mon
    day = date.tm_mday
    months.append(month)
    days.append(day)

# plot the PC phase diagram 
pmi.phase_diagram('RMM',PC1,PC2,time,months,days,'RMM_comp_phase','png')

# plot PC time series
pmi.pc_time_series('RMM',PC1,PC2,time,months,days,'RMM_time_series','png')


# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
from typing import Union
import math

from statistics import mean, median
import numpy as np
from itertools import groupby
from scipy.stats import norm

from metcalcpy.util.correlation import remove_none


def runs_test(x, alternative="two.sided", threshold='median') -> dict:
    """
    Wald-Wolfowitz Runs Test.
    Performs the Wald-Wolfowitz runs test of randomness for continuous data.
    Method used to compute the p-value is normal
    Mimics runs.test Rscript function

    :param x: a numeric vector containing the observations
    :param alternative: a character string with the alternative hypothesis.
        Must be one of "two.sided" (default), "left.sided" or "right.sided".
    :param threshold: the cut-point to transform the data into a dichotomous vector
        Must be one of "median" (default) or "mean".
    :return: Wald-Wolfowitz Runs Test results as a dictionary
        statistic - the value of the normalized statistic test
        p_value - the p-value of the test
        runs - the total number of runs
        mean_value - the mean value of the statistic test
        variance - the variance of the statistic test
    """
    result = {
        'statistic': None,
        'p_value': None,
        'runs': None,
        'mean_value ': None,
        'variance': None
    }
    if alternative != "two.sided" and alternative != "left.sided" and alternative != "right.sided":
        print(" runs_test must get a valid alternative")
        return result
    if len(x) == 0:
        return result

    x = remove_none(x)
    if threshold == 'median':
        x_threshold = median(x)
    elif threshold == "mean":
        x_threshold = mean(x)
    else:
        print('ERROR  incorrect threshold')
        x_threshold = None
    x = [elem for elem in x if elem != x_threshold]
    res = [i - x_threshold for i in x]
    s = np.sign(res)
    n1 = 0
    n2 = 0
    for num in s:
        if num > 0:
            n1 += 1
        elif num < 0:
            n2 += 1
    runs = [(k, sum(1 for i in g)) for k, g in groupby(s)]
    r1 = 0
    r2 = 0
    for run in runs:
        if run[0] == 1:
            r1 += 1
        elif run[0] == -1:
            r2 += 1
    n = n1 + n2
    mean_value = 1 + 2 * n1 * n2 / (n1 + n2)
    variance = 2 * n1 * n2 * (2 * n1 * n2 - n1 - n2) / (n * n * (n - 1))
    rr = r1 + r2
    pv = 0
    pv0 = norm.cdf((rr - mean_value) / math.sqrt(variance))
    if alternative == "two.sided":
        pv = 2 * min(pv0, 1 - pv0)
    if alternative == "left.sided":
        pv = pv0
    if alternative == "right.sided":
        pv = 1 - pv0

    result['statistic'] = (rr - mean_value) / math.sqrt(variance)
    result['p_value'] = pv
    result['runs'] = rr
    result['mean_value'] = mean_value
    result['variance'] = variance

    return result



"""Tests the operation of METcalcpy's event_equalize code."""
import itertools
import time
import re

import pandas as pd
import pytest

from metcalcpy.event_equalize import event_equalize
from metcalcpy.util.utils import represents_int
from metcalcpy.util.utils import equalize_axis_data
from metcalcpy import GROUP_SEPARATOR, DATE_TIME_REGEX

def test_event_equalize():
    """Tests event equalization."""

    indy_var = "fcst_lead"
    series_val = dict({'model': ["GFSDCF", "GFSRAW"]})
    fixed_vars_vals_input = dict({
        'fcst_thresh': dict({'fcst_thresh_4': ["<=20"]}),

    })

    fcst_var_val = dict({'TCDC': ["BASER"]})
    input_data_file = 'data/event_equalize_input.data'

    # read the input data file into a data frame
    input_data = pd.read_csv(input_data_file, header=[0], sep='\t')

    output_data = perform_event_equalize(fcst_var_val, fixed_vars_vals_input,
                                         indy_var, input_data, series_val)

    assert len(input_data) == 244
    assert len(output_data) == 164

    # test groups

    indy_var = "fcst_lead"
    series_val = dict({'model': ["CONTROL:GTS"],
                       'fcst_valid_beg': ['2010-06-01 00:00:00:2010-06-01 12:00:00:2010-06-02 00:00:00']})
    fixed_vars_vals_input = dict()

    fcst_var_val = dict({'TMP': ["ME"]})
    input_data_file = 'data/event_equalize_group_input.data'

    # read the input data file into a data frame
    input_data = pd.read_csv(input_data_file, header=[0], sep='\t')
    output_data = perform_event_equalize(fcst_var_val, fixed_vars_vals_input,
                                         indy_var, input_data, series_val)

    assert len(input_data) == 216
    assert len(output_data) == 216


def perform_event_equalize(fcst_var_val, fixed_vars_vals_input, indy_var, input_data, series_val):
    cl_step = 0.05
    bool_multi = False
    start_all = time.time()

    # sort the dataset by init time, lead time and independent variable
    input_data.sort_values(by=['fcst_init_beg', 'fcst_lead', indy_var])
    fix_vars = []
    fix_vals = []
    output_data = pd.DataFrame()
    # list all fixed variables
    if fixed_vars_vals_input:
        for key, value in fixed_vars_vals_input.items():
            fix_vars.append(key)
            fix_vals.append(list(value.values()))
    # permute fix vals
    fix_vals_permuted = list(itertools.chain.from_iterable(fix_vals))
    # perform EE for each forecast variable
    for fcst_var, fcst_var_stats in fcst_var_val.items():
        for fcst_var_stat in fcst_var_stats:
            for series_var, series_var_vals in series_val.items():
                # ungroup series value
                series_var_vals_no_group = []
                for val in series_var_vals:
                    split_val = re.findall(DATE_TIME_REGEX, val)
                    if len(split_val) == 0:
                        split_val = val.split(GROUP_SEPARATOR)
                    series_var_vals_no_group.extend(split_val)

                series_data = input_data[(input_data['fcst_var'] == fcst_var)
                                         & (input_data["stat_name"] == fcst_var_stat)
                                         & (input_data[series_var].isin(series_var_vals_no_group))]
                start = time.time()
                series_data = \
                    event_equalize(series_data, indy_var, series_val, fix_vars,
                                   fix_vals_permuted, True, bool_multi)
                end = time.time()
                print("one EE:" + str(end - start))

                # append EE data to result
                if output_data.empty:
                    output_data = series_data
                else:
                    output_data.append(series_data)
    end_all = time.time()
    print("total :" + str(end_all - start_all))
    return output_data


@pytest.fixture
def settings():
    """Initialise values for testing.

    Returns:
        dictionary with values of different type
    """
    settings_dict = dict()
    settings_dict['int'] = 8
    settings_dict['str'] = 'str'
    settings_dict['date'] = '2019-09-05 03:00:00'
    settings_dict['double'] = 34.9
    settings_dict['int_n'] = -700
    return settings_dict

def test_represents_int_not_int(settings):
    """Tests that this fails to cast a type that cannot be cast into an int
        Args:
            settings: dictionary with values of different type
    """
    assert not represents_int(settings['str'])
    assert represents_int(settings['int'])
    assert represents_int(settings['int_n'])
    assert not represents_int(settings['double'])
    assert not represents_int(settings['date'])

@pytest.fixture
def settings_no_fix_vals():
    """

     Returns:
          settings_dict: a dictionary corresponding to the params arg for equalize_axis_data()
    """
    settings_dict = dict()
    settings_dict['fix_val_keys'] = []
    settings_dict['fcst_var_val_1'] = {}
    settings_dict['fcst_var_val_2'] = {}
    settings_dict['fix_vals_permuted'] = {}
    settings_dict['series_val_1'] = {}
    settings_dict['indy_var'] = {}
    settings_dict['line_type'] = None

    return settings_dict

def test_equalize_axis_data_no_fix_val(settings_no_fix_vals):
    '''
       conditions that lead to an empty data frame with no 'equalize' column following event equalization should
       return an empty data frame (instead of an error due to trying to clean up the 'equalize' column before
       returning the event equalization data frame).
    '''

    print("Testing equalize_axis_data with ROC CTC threshold data...")
    input_file_list = ["data/ROC_CTC_thresh.data", "data/ROC_CTC.data"]
    for input_file in input_file_list:
        cur_df = pd.read_csv(input_file, sep='\t')
        fix_vals_keys = []
        fix_vals_permuted_list = []

        ee_df = equalize_axis_data(fix_vals_keys, fix_vals_permuted_list, settings_no_fix_vals, cur_df, axis='1')

        assert ee_df.shape[0] == 0


@pytest.fixture
def settings_no_fcst_var_vals():
    """

     Returns:
          settings_dict: a dictionary corresponding to the params arg for equalize_axis_data()
    """
    settings_dict = dict()
    settings_dict['fix_val_keys'] = []
    # No fcst_var_val in 'config'
    # settings_dict['fcst_var_val_1'] = {}
    # settings_dict['fcst_var_val_2'] = {}
    settings_dict['fix_vals_permuted'] = {}
    settings_dict['series_val_1'] = {}
    settings_dict['indy_var'] = {}
    settings_dict['line_type'] = None

    return settings_dict

def test_equalize_axis_data_no_fcst_var(settings_no_fcst_var_vals):
    '''
       conditions that lead to an empty data frame from event equalization should
       return an empty data frame instead of an error due to trying to clean up the 'equalize' column before
       returning the event equalization data frame.
    '''

    print("Testing equalize_axis_data with ROC CTC threshold data...")
    input_file_list = ["data/ROC_CTC_thresh.data", "data/ROC_CTC.data"]
    for input_file in input_file_list:
        cur_df = pd.read_csv(input_file, sep='\t')
        fix_vals_keys = []
        fix_vals_permuted_list = []

        ee_df = equalize_axis_data(fix_vals_keys, fix_vals_permuted_list, settings_no_fcst_var_vals, cur_df, axis='1')
        assert ee_df.shape[0] == 0


if __name__ == "__main__":
    test_event_equalize()



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ** University of Illinois, Urbana-Champaign
 # ============================*
 
 
 
import os
import numpy as np
from pylab import *
from sklearn.cluster import KMeans
import scipy
import netCDF4 as nc4
from scipy import stats,signal
from numpy import ones,vstack
from numpy.linalg import lstsq
from eofs.standard import Eof


class WeatherRegimeCalculation():
    """Contains the programs to perform a Weather Regime Analysis
    """
    def __init__(self,label):

        self.wrnum = int(os.environ.get(label+'_WR_NUMBER',6))
        self.numi = int(os.environ.get(label+'_NUM_CLUSTERS',20))
        self.NUMPCS = int(os.environ.get(label+'_NUM_PCS',10))
        self.wr_tstep = int(os.environ.get(label+'_WR_FREQ',7))
        self.wr_outfile_type = os.environ.get(label+'_WR_OUTPUT_FILE_TYPE','text')
        self.wr_outfile_dir = os.environ.get('WR_OUTPUT_FILE_DIR',os.environ['SCRIPT_OUTPUT_BASE'])
        self.wr_outfile = os.environ.get(label+'_WR_OUTPUT_FILE',label+'_WeatherRegime')


    def get_cluster_fraction(self, m, label):
        return (m.labels_==label).sum()/(m.labels_.size*1.0)


    def weights_detrend(self,lats,lons,indata):

        arr_shape = indata.shape

        ##Set up weight array
        cos = lats * np.pi / 180.0
        way = np.cos(cos)
        if len(lats.shape) == 1:
            weightf = np.repeat(way[:,np.newaxis],len(lons),axis=1)
        else:
            weightf = way

        #Remove trend and seasonal cycle
        atemp = np.zeros(arr_shape)
        for i in np.arange(0,len(indata[0,:,0,0]),1):
            atemp[:,i] = signal.detrend(atemp[:,i],axis=0)
            atemp[:,i] = (indata[:,i] - np.nanmean(indata[:,i],axis=0)) * weightf

        a = atemp
        atemp=0

        #Reshape into time X space
        a1 = np.reshape(a,[len(a[:,0,0,0])*len(a[0,:,0,0]),len(lons)*len(lats)])

        return a,a1


    def run_elbow(self,a1):

        k=KMeans(n_clusters=self.wrnum, random_state=0)  #Initilize cluster centers

        #Calculate sum of squared distances for clusters 1-15
        kind = np.arange(1,self.numi,1)
        Sum_of_squared_distances = []
        K = range(1,self.numi)
        for k in K:
            km = KMeans(n_clusters=k)
            km = km.fit(a1)
            Sum_of_squared_distances.append(km.inertia_)

        # Calculate the bend of elbow
        points = [(K[0],Sum_of_squared_distances[0]),(K[-1],Sum_of_squared_distances[-1])]
        x_coords, y_coords = zip(*points)
        A = vstack([x_coords,ones(len(x_coords))]).T
        m, c = lstsq(A, y_coords,rcond=None)[0]
        line = m*kind + c
        curve = Sum_of_squared_distances
        curve=np.array(curve)*10**-10
        line = line*10**-10

        d=[]
        for i in np.arange(0,self.numi-1,1):
            p1=np.array([K[0],curve[0]])
            p2=np.array([K[-1],curve[-1]])
            p3=np.array([K[i],curve[i]])
            d=np.append(d,np.cross(p2-p1,p3-p1)/np.linalg.norm(p2-p1))

        mi = np.where(d==d.min())[0]
        print('Optimal Cluster # = '+str(mi+1)+'')

        return K,d,mi,line,curve


    def Calc_EOF(self,eofin):

        #Remove trend and seasonal cycle
        for d in np.arange(0,len(eofin[0,:,0,0]),1):
            eofin[:,d] = signal.detrend(eofin[:,d],axis=0)
            eofin[:,d] = eofin[:,d] - np.nanmean(eofin[:,d],axis=0)

        #Reshape into time X space
        arr_shape = eofin.shape
        arrdims = len(arr_shape)
        eofin = np.reshape(eofin,(np.prod(arr_shape[0:arrdims-2]),arr_shape[arrdims-2]*arr_shape[arrdims-1]))

        # Use EOF solver and get PCs and EOFs
        solver = Eof(eofin,center=False)
        pc = solver.pcs(npcs=self.NUMPCS,pcscaling=1)
        eof = solver.eofsAsCovariance(neofs=self.NUMPCS,pcscaling=1)
        eof = np.reshape(eof,(self.NUMPCS,arr_shape[arrdims-2],arr_shape[arrdims-1]))

        #Get variance fractions
        variance_fractions = solver.varianceFraction(neigs=self.NUMPCS) * 100
        print(variance_fractions)

        return eof, pc, self.wrnum, variance_fractions


    def reconstruct_heights(self,eof,pc,reshape_arr):

        rssize = len(reshape_arr)
        eofshape = eof.shape
        eosize = len(eofshape)

        #reconstruction. If NUMPCS=nt, then a1=a0
        eofs=np.reshape(eof,(self.NUMPCS,eofshape[eosize-2]*eofshape[eosize-1]))
        a1=np.matmul(pc,eofs)

        return a1


    def run_K_means(self,a1,timedict,arr_shape):

        arrdims = len(arr_shape)

        k=KMeans(n_clusters=self.wrnum, random_state=0)

        #fit the K-means algorithm to the data
        f=k.fit(a1)

        #Obtain the cluster anomalies
        y=f.cluster_centers_

        #Obtain cluster labels for each day [Reshape to Year,day]
        wr = f.labels_
        wr = np.reshape(wr,arr_shape[0:arrdims-2])

        yf = np.reshape(y,[self.wrnum,arr_shape[arrdims-2],arr_shape[arrdims-1]]) # reshape cluster anomalies to latlon

        #Get frequency of occurence for each cluster
        perc=np.zeros(self.wrnum)
        for ii in np.arange(0,self.wrnum,1):
            perc[ii] = self.get_cluster_fraction(f,ii)

        #Sort % from low to high
        ii = np.argsort(perc)
        print(perc[ii])

        #Reorder
        perc = perc[ii]
        input=yf[ii]
        ii = ii[::-1]

        #Reorder from max to min and relabel
        wrc = wr*1.0/1.0
        for i in np.arange(0,self.wrnum,1):
            wrc[wr==ii[i]] = i+1

        perc = perc[::-1]
        input = input[::-1]

        return input, self.wrnum, perc, wrc


    def write_K_means_file(self,timedict,wrc):
        #Save Label data [YR,DAY]
        # Make some conversions first
        wrc_shape = wrc.shape
        len1d =  wrc.size
        valid_time_1d = np.reshape(timedict['valid'],len1d)
        yr_1d = []
        mth_1d = []
        day_1d = []
        for vt1 in valid_time_1d:
           yr_1d.append(vt1[0:4])
           mth_1d.append(vt1[4:6])
           day_1d.append(vt1[6:8])
        wrc_1d = np.reshape(wrc,len1d)

        # netcdf file
        if self.wr_outfile_type=='netcdf':
            wr_full_outfile = os.path.join(self.wr_outfile_dir,self.wr_outfile+'.nc')

            if os.path.isfile(wr_full_outfile):
                os.remove(wr_full_outfile)

            # Create CF compliant time unit
            rdate = datetime.datetime(int(yr_1d[0]),int(mth_1d[0]),int(day_1d[0]),0,0,0)
            cf_diffdays = np.zeros(len(yr_1d))
            ymd_arr = np.empty(len(yr_1d))
            for dd in range(len(yr_1d)):
                loopdate = datetime.datetime(int(yr_1d[dd]),int(mth_1d[dd]),int(day_1d[dd]),0,0,0)
                cf_diffdays[dd] = (loopdate - rdate).days
                ymd_arr[dd] = yr_1d[dd]+mth_1d[dd]+day_1d[dd]

            nc = nc4.Dataset(wr_full_outfile, 'w')
            nc.createDimension('time', len(mth_1d))
            nc.Conventions = "CF-1.7"
            nc.title = "Weather Regime Classification"
            nc.institution = "NCAR DTCenter"
            nc.source = "Weather Regime METplus use-case"

            ncti = nc.createVariable('time','d',('time'))
            nc.variables['time'].long_name = "time"
            nc.variables['time'].standard_name = "time"
            nc.variables['time'].units = "days since "+rdate.strftime('%Y-%m-%d %H:%M:%S')
            nc.variables['time'].calendar = "gregorian"

            ncdate = nc.createVariable('date','i',('time'))
            nc.variables['date'].long_name = "date YYYYMMDD"
       
            ncnum = nc.createVariable('wrnum','i',('time'),fill_value=-9999.0)
            nc.variables['wrnum'].long_name = "weather_regime_number"

            ncti[:] = cf_diffdays
            ncdate[:] = ymd_arr
            ncnum[:] = wrc_1d
            nc.close()

        # text file
        if self.wr_outfile_type=='text':
           wr_full_outfile = os.path.join(self.wr_outfile_dir,self.wr_outfile+'.txt')

           if os.path.isfile(wr_full_outfile):
                os.remove(wr_full_outfile)

           otdata = np.array([yr_1d, mth_1d, day_1d, wrc_1d])
           otdata = otdata.T

           with open(wr_full_outfile, 'w+') as datafile_id:
               np.savetxt(datafile_id, otdata, fmt=['%6s','%3s','%4s','%6s'], header='Year Month Day WeatherRegime')


    def compute_wr_freq(self, WR):

        ######## Simple Count ##########
        WRfreq = np.zeros((self.wrnum,len(WR[:,0]),len(WR[0,:])-self.wr_tstep+1))

        for yy in np.arange(0,len(WRfreq[0,:,0]),1):
            d1=0;d2=self.wr_tstep
            for dd in np.arange(len(WRfreq[0,0,:])):
                temp = WR[yy,d1:d2]
                for ww in np.arange(self.wrnum):
                    WRfreq[ww,yy,dd] = len(np.where(temp==ww+1)[0])
                d1=d1+1;d2=d2+1

        dlen_plot = len(WRfreq[0,0,:])

        return WRfreq, dlen_plot, self.wr_tstep



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
# -*- coding: utf-8 -*-
"""
Piecewise linear function class.

"""

import numpy as np

__author__ = 'Bill Campbell (NRL)'
__version__ = '0.1.0'


class IncompatibleLengths(Exception):
    """Custom exception for PiecewiseLinear input checking."""


class UnsortedArray(Exception):
    """Custom exception for PiecewiseLinear input checking."""


class PiecewiseLinear():
    """
    Defines a piecewise linear function with a given domain and range.

    Xdomain is a numpy array of knot locations. Yrange is a numpy array.
    """

    def __init__(self, x_domain, y_range, xunits='feet',
                 left=np.nan, right=np.nan, name=""):
        len_x = len(x_domain)
        if len_x < 2:
            raise IncompatibleLengths('Length of xdomain must be at least 2.')
        if np.any(np.diff(x_domain)) < 0:
            print("X_domain (in {}) is {}".format(xunits, x_domain))
            raise UnsortedArray('Xdomain must be sorted in ascending order.')
        len_y = len(y_range)
        if len_x != len_y:
            raise IncompatibleLengths('X_domain and Y_range must have same ' +
                                      'length.\n Use left and right to set ' +
                                      'value for points outside the x_domain\n')
        self.x_domain = np.array(x_domain)
        self.y_range = np.array(y_range)
        self.xunits = xunits
        self.left = left
        self.right = right
        self.name = name

    def get_ymax(self):
        """Find maximum of envelope function"""
        return np.max(self.y_range)

    def values(self, xinput):
        """
        Evaluate piecewise linear function for the set of points in xinput.

        xinput is a set of points inside xdomain.
        """
        if not isinstance(xinput, np.ndarray):
            xinput = np.array(xinput)
        xin_shape = np.shape(xinput)
        # Treat xinput as one-dimensional for use by np.interp
        xview = xinput.ravel()
        yflat = np.interp(xview, self.x_domain, self.y_range,
                          left=self.left, right=self.right)
        # Restore output to same shape as input
        youtput = np.reshape(yflat, xin_shape)

        return youtput

if __name__ == "__main__":
    pass



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: mcts_statistics.py
"""
import warnings
import numpy as np
from metcalcpy.util.utils import round_half_up, sum_column_data_by_name, PRECISION

__author__ = 'Tatiana Burek'


def calculate_mcts_hss_ec(input_data, columns_names):
    """Performs calculation of HSS_EC - a skill score based on Accuracy,

        Args:
            input_data: 2-dimensional numpy array with data for the calculation
                1st dimension - the row of data frame
                2nd dimension - the column of data frame
            columns_names: names of the columns for the 2nd dimension as Numpy array

        Returns:
            calculated HSS_EC as float
            or None if some of the data values are missing or invalid
    """
    warnings.filterwarnings('error')

    try:
        row = input_data[0].copy()
        n_cat = row[np.where(columns_names == 'n_cat')[0][0]]
        ec_value = row[np.where(columns_names == 'ec_value')[0][0]]

        # aggregate all fi_oj in one row
        for index in range(n_cat * n_cat):
            column_name = 'fi_oj_' + str(index)
            row[np.where(columns_names == column_name)[0][0]] = \
                sum_column_data_by_name(input_data, columns_names, column_name)

        # init contingency table
        cont_table = [[0] * n_cat for _ in range(n_cat)]

        # fill contingency table
        for index in range(n_cat * n_cat):
            i_value = row[np.where(columns_names == 'i_value_' + str(index))[0][0]]
            j_value = row[np.where(columns_names == 'j_value_' + str(index))[0][0]]
            fi_oj = row[np.where(columns_names == 'fi_oj_' + str(index))[0][0]]
            cont_table[i_value - 1][j_value - 1] = fi_oj

        # calculate the sum of the counts on the diagonal and
        # the sum of the counts across the whole MCTC table
        diag_count = sum([cont_table[i][j] for i in range(n_cat) for j in range(n_cat) if i == j])
        sum_all = sum(sum(cont_table, []))
        result = (diag_count - (ec_value * sum_all)) / (sum_all - (ec_value * sum_all))
        result = round_half_up(result, PRECISION)
    except (TypeError, ZeroDivisionError, Warning, ValueError):
        result = None
    warnings.filterwarnings('ignore')
    return result



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: vertical_interp.py

A python script to vertically interpolate fields
between grids with pressure or height vertical coordinates.

Currently only pressure to height conversion is implemented.
Currently only linear interpolation is implemented.

Version  Date
0.1.0    2020/09/01  David Fillmore  Initial version

Known Issues:
    xarray does not propagate coordinate attributes to interpolation field
    need to handle unknown units

Needed Enhancements:
    multiple input files on command line or config file
    config file definition of user defined units
    config file option for interpolation weights
        above (below) the highest (lowest) coordinate surface
"""

__author__ = 'David Fillmore'
__version__ = '0.1.0'

import metpy.units

"""
Import standard modules
"""
import os
import sys
import argparse
import logging
import yaml
from datetime import datetime
import numpy as np
import xarray as xr  # http://xarray.pydata.org/
import netCDF4 as nc

"""
Import Pint and MetPy modules
    https://unidata.github.io/MetPy/
"""
import pint
import metpy.calc as calc
import metpy.constants as constants

ureg = pint.UnitRegistry()

def vertical_interp(fieldname, config,
    coordinate_surfaces, field):
    """
    Interpolate field onto coordinate surfaces.
    Linear interpolation is the only method currently implemented.

    Arguments:
        fieldname (string) : short name of field
        config (dictionary) : configuration parameters
        coordinate_surfaces (DataArray): coordinate surfaces
        field (DataArray): field

    Returns:
        field_interp (DataArray): Interpolated field
    """
    logging.info(fieldname)
    logging.debug(field.attrs)

    """
    Vertical coordinates
    """
    lev_dim = config['vertical_dim_name']
    vertical_coord = field.coords[lev_dim]
    nlev = len(vertical_coord)
    vertical_indices = np.arange(nlev)

    """
    Vertical coordinates on which to interpolate
    """
    vertical_levels = np.array(config['vertical_levels'], dtype=field.dtype)
    nlev_interp = len(vertical_levels)
    logging.debug(vertical_levels)
    logging.debug(coordinate_surfaces.shape)

    """
    Setup interpolated field shape and coordinates
    """
    dims_interp = list(field.dims)
    i_lev_dim = dims_interp.index(lev_dim)
    logging.debug(('dims_interp:', dims_interp))
    shape_interp = list(field.shape)
    shape_interp[i_lev_dim] = nlev_interp
    shape_interp = tuple(shape_interp)
    logging.debug(shape_interp)
    coord_arrays_interp = [field.coords[dim].data for dim in dims_interp]
    dims_interp[i_lev_dim] = 'lev'
    """
    coord_arrays_interp[i_lev_dim] \
        = xr.DataArray(vertical_levels,
                       attrs={'units': config['vertical_level_units']})
    """
    coord_arrays_interp[i_lev_dim] = vertical_levels
    coords_interp = list(zip(dims_interp, coord_arrays_interp))
    logging.debug('\n\n')
    for coord_interp in coords_interp:
        logging.debug(coord_interp)
        logging.debug('\n\n')

    """
    Setup dimensions and shape for a vertical slice
    """
    dims_slice = list(field.dims)
    dims_slice.pop(i_lev_dim)
    dims_slice = tuple(dims_slice)
    logging.debug(dims_slice)
    shape_slice = list(field.shape)
    shape_slice.pop(i_lev_dim)
    shape_slice = tuple(shape_slice)
    logging.debug(shape_slice)
    field_slice = field.drop(lev_dim)
    coords_slice = field_slice.coords
    logging.debug(coords_slice)

    """
    Initialize interpolated field
    """
    field_interp = xr.DataArray(
        np.zeros(shape_interp),
        dims=dims_interp,
        coords=coords_interp,
        attrs=field.attrs)
    # coordinate attributes lost in constructor above

    # length unit conversion
    try:
        logging.debug(coordinate_surfaces.attrs['units'])
        logging.debug(config['vertical_level_units'])
        length_convert = float((ureg.Quantity(1, config['vertical_level_units'])
                       / ureg.Quantity(1, coordinate_surfaces.attrs['units'])).to_base_units())
    except pint.errors.UndefinedUnitError:
        logging.warning('Unknown unit:' + coordinate_surfaces.attrs['units'])
        length_convert = 1
    logging.debug(length_convert)

    for k_interp, eta in zip(range(nlev_interp), vertical_levels):
        """
        Compute interpolation weights
        """

        logging.debug((k_interp, eta))

        weights = xr.DataArray(
            np.zeros(field.shape),
            dims=field.dims,
            coords=field.coords)

        distances = eta - length_convert * coordinate_surfaces
        above = distances < 0
        below = distances > 0

        """
        Compute weights for linear interpolation
        """
        # where bottom most layer is above eta
        layer_above = above.loc[{lev_dim: vertical_coord[0]}]
        weights.loc[{lev_dim: vertical_coord[0]}] \
            = xr.where(layer_above, 1, 0)

        for k in vertical_indices[1:]:
            layer_above = above.loc[{lev_dim: vertical_coord[k]}]
            layer_below = below.loc[{lev_dim: vertical_coord[k - 1]}]
            # where upper layer is above and lower layer is below eta
            mask = np.logical_and(layer_above, layer_below)

            # w(z_1) = 1 - (z_1 - eta) / (z_1 - z_0) =  (eta - z_0) / (z_1 - z_0)
            weight_above = distances.loc[{lev_dim: vertical_coord[k - 1]}] \
                / ((coordinate_surfaces.loc[{lev_dim: vertical_coord[k]}]
                    - coordinate_surfaces.loc[{lev_dim: vertical_coord[k - 1]}])
                    * length_convert)
            # w(z_0) = 1 - w_1 = (z_1 - eta) / (z_1 - z_0)
            weight_below = 1 - weight_above

            weights.loc[{lev_dim: vertical_coord[k]}] \
                = xr.where(mask, weight_above,
                           weights.loc[{lev_dim: vertical_coord[k]}])
            weights.loc[{lev_dim: vertical_coord[k - 1]}] \
                = xr.where(mask, weight_below,
                           weights.loc[{lev_dim: vertical_coord[k - 1]}])

        # where top most layer is below eta
        layer_below = below.loc[{lev_dim: vertical_coord[nlev - 1]}]
        weights.loc[{lev_dim: vertical_coord[nlev - 1]}] \
            = xr.where(layer_below, 1, 0)

        """
        Compute weighted sum
        """
        field_slice = xr.DataArray(
            np.zeros(shape_slice),
            dims=dims_slice,
            coords=coords_slice,
            attrs=field.attrs)

        counts = xr.DataArray(
            np.zeros(shape_slice),
            dims=dims_slice,
            coords=coords_slice,
            attrs=field.attrs)

        for k in vertical_indices:
            weights_k = weights.loc[{lev_dim: vertical_coord[k]}]
            field_k = field.loc[{lev_dim: vertical_coord[k]}]
            mask = weights_k > 0
            field_slice = xr.where(mask,
                                   field_slice + weights_k * field_k, field_slice)
            counts = xr.where(mask, counts + 1, counts)

        mask = counts > 0
        field_slice = xr.where(mask, field_slice, np.nan)

        # field_interp[dict(isobaricInhPa=k_interp)] = field_slice
        # field_interp[{lev_dim: k_interp}] = field_slice
        field_interp[dict(lev=k_interp)] = field_slice

        """
        Write fields for debugging
        """
        if logging.root.level == logging.DEBUG:
            ds_debug = xr.Dataset(
                {'distances': distances,
                 'weights': weights,
                 fieldname: field_slice})
            debugfile = os.path.join(args.datadir, 'Debug', 'vertical_interp_debug_'
                                     + fieldname + '_' + str(int(eta)) + '.nc')
            try:
                ds_debug.to_netcdf(debugfile)
            except:
                ds_nc = nc.Dataset(debugfile, 'w')
                write_dataset(ds_debug, ds_nc)
                ds_nc.close()

    return field_interp, coords_interp


def height_from_pressure(config,
    surface_geopotential, surface_pressure,
    temperature, relative_humidity):
    """
    Compute height coordinate surfaces as a function of pressure.

    Arguments:
        config (dictionary) : configuration parameters
        surface_geopotential (DataArray) : surface geopotential
        surface_pressure (DataArray) : surface pressure
        temperature (DataArray) : temperature
        relative_humidity (DataArray) : relative humidity

    Returns:
        layer_height (DataArray) : layer height
    """

    logging.info('pressure to height conversion')

    """
    Compute surface geopotential height from geopotential
    """
    surface_height = xr.DataArray(
        surface_geopotential / constants.earth_gravity.to_base_units().magnitude,
        dims=surface_geopotential.dims,
        coords=surface_geopotential.coords,
        attrs={'long_name': 'surface geopotential height',
               'units': 'meter'})

    """
    Get pressure coordinates
    """
    logging.debug(temperature.coords)
    logging.debug(temperature.shape)
    lev_dim = config['vertical_dim_name']
    pressure_coord = temperature.coords[lev_dim]
    nlev = len(pressure_coord)
    pressure_indices = np.arange(nlev)

    """
    Create pressure field
    """
    pressure = xr.DataArray(
        np.empty(temperature.shape),
        dims=temperature.dims,
        coords=temperature.coords,
        attrs={'long_name': 'pressure',
               'units': pressure_coord.attrs['units']})
    for p in pressure_coord:
        pressure.loc[{lev_dim: p}] = p

    """
    Compute mixing ratio
    """
    mixing_ratio \
        = xr.DataArray(
            calc.mixing_ratio_from_relative_humidity(
                pressure, temperature, relative_humidity),
            dims=temperature.dims,
            coords=temperature.coords,
            attrs={'long_name': 'mixing ratio'})

    """
    Compute virtual temperature
    """
    virtual_temperature \
        = xr.DataArray(
            calc.virtual_temperature(temperature, mixing_ratio),
            dims=temperature.dims,
            coords=temperature.coords,
            attrs={'long_name': 'virtual temperature',
                   'units': temperature.attrs['units']})

    """
    Compute layer thickness
    Z_2 - Z_1 = (R_d / g) <T_v> log(p_1 / p_2)
    R_d / g = dry_air_gas_constant / earth_gravity
    <T_v> = integral_p_2^p_1 T_v(p) (dp / p) / log(p_1 / p_2)
    """
    gas_constant_gravity_ratio \
        = (constants.dry_air_gas_constant
        / constants.earth_gravity).to_base_units()
    logging.debug(gas_constant_gravity_ratio)
    logging.debug(surface_pressure.attrs['units'])
    logging.debug(pressure.attrs['units'])

    # pressure unit conversion
    pressure_convert = float((ureg.Quantity(1, surface_pressure.attrs['units'])
        / ureg.Quantity(1, pressure.attrs['units'])).to_base_units())
    logging.debug(pressure_convert)

    layer_thickness = xr.DataArray(
        np.empty(temperature.shape),
        dims=temperature.dims,
        coords=temperature.coords,
        attrs={'long_name': 'layer thickness',
               'units': 'meter'})

    surface_mask = xr.DataArray(
        np.empty(temperature.shape, dtype=bool),
        dims=temperature.dims,
        coords=temperature.coords,
        attrs={'long_name': 'surface mask'})

    layer_height = xr.DataArray(
        np.empty(temperature.shape),
        dims=temperature.dims,
        coords=temperature.coords,
        attrs={'long_name': 'layer height',
               'units': 'meter'})

    layer_thickness.loc[{lev_dim: pressure_coord[0]}] \
        = gas_constant_gravity_ratio \
        * virtual_temperature.loc[{lev_dim: pressure_coord[0]}] \
        * np.log(pressure_convert * surface_pressure
                 / pressure.loc[{lev_dim: pressure_coord[0]}])

    for k in pressure_indices[1:]:
        # logging.debug(k)
        layer_thickness.loc[{lev_dim:pressure_coord[k]}] \
            = gas_constant_gravity_ratio \
            * virtual_temperature.loc[{lev_dim: pressure_coord[k]}] \
            * np.log(pressure.loc[{lev_dim: pressure_coord[k - 1]}] \
            / pressure.loc[{lev_dim: pressure_coord[k]}])

    layer_thickness = layer_thickness.fillna(0)
    layer_thickness = layer_thickness.clip(min=0)

    """
    Compute layer height and surface mask
        The surface mask value value is true if the surface pressure
        is between the lower layer and upper (current) layers.
    """
    surface_mask.loc[{lev_dim: pressure_coord[0]}] \
        = pressure.loc[{lev_dim: pressure_coord[0]}] \
        < pressure_convert * surface_pressure
    layer_height.loc[{lev_dim: pressure_coord[0]}] \
        = xr.where(surface_mask.loc[{lev_dim: pressure_coord[0]}],
                   surface_height + layer_thickness.loc[{lev_dim: pressure_coord[0]}],
                   np.nan)

    for k in pressure_indices[1:]:

        surface_mask.loc[{lev_dim: pressure_coord[k]}] \
            = np.logical_and(
            pressure.loc[{lev_dim: pressure_coord[k - 1]}]
            > pressure_convert * surface_pressure,
            pressure.loc[{lev_dim: pressure_coord[k]}]
            < pressure_convert * surface_pressure)

        layer_height.loc[{lev_dim: pressure_coord[k]}] \
            = xr.where(surface_mask.loc[{lev_dim: pressure_coord[k]}],
                surface_height + layer_thickness.loc[{lev_dim: pressure_coord[k]}],
                layer_height.loc[{lev_dim: pressure_coord[k - 1]}]
                + layer_thickness.loc[{lev_dim: pressure_coord[k]}])

    """
    Write fields for debugging
    Approximate layer heights
        800 hPa ~ 2 km, 500 hPa ~ 6 km, 200 hPa ~ 12 km
    """
    if logging.root.level == logging.DEBUG:
        ds_debug = xr.Dataset(
            {'surface_geopotential': surface_geopotential,
             'surface_height': surface_height,
             'surface_pressure': surface_pressure,
             'surface_mask': surface_mask,
             'layer_pressure': pressure,
             'mixing_ratio': mixing_ratio,
             'virtual_temperature': virtual_temperature,
             'layer_thickness': layer_thickness,
             'layer_height': layer_height})
        debugfile = os.path.join(args.datadir, 'Debug',
            'height_from_pressure_debug.nc')
        
        # create the Debug directory if it doesn't already exist
        debug_path = os.path.join(args.datadir, 'Debug')
        if not os.path.exists(debug_path):
            os.makedirs(debug_path)
        try:
            ds_debug.to_netcdf(debugfile)
        except:
            ds_nc = nc.Dataset(debugfile, 'w')
            write_dataset(ds_debug, ds_nc)
            ds_nc.close() 

    return layer_height


def read_required_fields(config, ds):
    """
    Read required fields.

    Arguments:
        config (dictionary) : configuration parameters
        ds (DataSet) : xarray dataset

    Returns:
        surface_geopotential (DataArray) : surface geopotential
        surface_pressure (DataArray) : surface pressure
        temperature (DataArray) : temperature
        relative_humidity (DataArray) : relative humidity
    """
    surface_pressure \
        = ds[config['surface_pressure_name']]

    if config['zero_surface_geopotential']:
        surface_geopotential = xr.zeros_like(surface_pressure)
    else:
        surface_geopotential \
            = ds[config['surface_geopotential_name']]

    temperature \
        = ds[config['temperature_name']]
    relative_humidity \
        = ds[config['relative_humidity_name']]
    return surface_geopotential, surface_pressure, \
        temperature, relative_humidity


def write_dataset(ds, ds_nc, coords_interp=None,
    forecast_reference_time=None, create_time_dim=False):
    """
    Write xarray Dataset to NetCDF file
    """
    coord_vars = {}
    for dim in ds.dims:
        logging.info('Creating dimension ' + dim)
        ds_nc.createDimension(dim, len(ds.coords[dim]))
        dtype = ds.coords[dim].dtype
        if dtype not in ['uint32', 'uint64', 'int32', 'int64', 'float32', 'float64']:
            dtype = 'float64'
        coord = ds_nc.createVariable(
            dim, dtype, (dim))
        coord_vars[dim] = coord
        if dim != 'time':
            coord[:] = ds.coords[dim].values
        else:
            dt_array = [datetime.utcfromtimestamp(dt.astype('O')/1e9)
                        for dt in ds.coords[dim].values]
            t_array = np.array([(dt - dt_array[0]).total_seconds()
                                for dt in dt_array], dtype=np.float64)
            coord[:] = t_array

    if create_time_dim:
        ds_nc.createDimension('valid_time', 1)
        time_coord = ds_nc.createVariable(
            'valid_time', 'float64', ('valid_time'))
        dt_valid = datetime.utcfromtimestamp(
            ds['valid_time'].astype('O')/1e9)
        dt_init = datetime.utcfromtimestamp(
            ds['init_time'].astype('O')/1e9)
        time_coord[:] = (dt_valid - dt_init).total_seconds()
        time_coord.long_name = 'valid_time'
        time_coord.units = 'seconds since ' + str(dt_init)

    if coords_interp is not None:
        for dim, coord_array in coords_interp:
            logging.info('Setting coordinate attributes for ' + dim)
            """
            for attr in coord_array.attrs:
                setattr(coord_vars[dim], attr, coord_array.attrs[attr])
            """

    for field in ds:
        logging.debug('Creating variable ' + field)
        dtype = ds[field].dtype
        if dtype not in ['uint32', 'uint64', 'int32', 'int64', 'float32', 'float64']:
            dtype = 'uint64'
        if create_time_dim:
            dims_with_time = list(ds[field].dims)
            dims_with_time.insert(0, 'valid_time')
            if field != 'valid_time':
                var = ds_nc.createVariable(
                    field, dtype, tuple(dims_with_time))
                var[:] = ds[field].values
        else:
            var = ds_nc.createVariable(
                field, dtype, ds[field].dims)
            var[:] = ds[field].values

        for attr in ds[field].attrs:
            logging.debug((attr, ds[field].attrs[attr]))
            setattr(var, attr, ds[field].attrs[attr])

    for attr in ds.attrs:
        logging.debug((attr, ds.attrs[attr]))
        setattr(ds_nc, attr, ds.attrs[attr])

    if forecast_reference_time is not None:
        # setattr(ds_nc,
        #     'forecast_reference_time', forecast_reference_time)
        ds_nc.createDimension('forecast_reference_time', 1)
        ref_time_coord = ds_nc.createVariable(
            'forecast_reference_time', 'float64',
            ('forecast_reference_time'))
        ref_time_coord.long_name = 'forecast_reference_time'
        ref_time_coord.standard_name = 'forecast_reference_time'
        ref_time_coord.units = 'seconds since 1970-01-01 00:00'
        yyyymmddhh_str = str(forecast_reference_time)
        yyyy = int(yyyymmddhh_str[0:4])
        mm = int(yyyymmddhh_str[4:6])
        dd = int(yyyymmddhh_str[6:8])
        hh = int(yyyymmddhh_str[8:10])
        ref_time_obj = datetime(yyyy, mm, dd, hh)
        ref_time_coord[:] = ref_time_obj.timestamp()


if __name__ == '__main__':
    """
    Parse command line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--datadir', type=str,
        default=os.getenv('DATA_DIR'),
        help='top-level data directory (default $DATA_DIR)')
    parser.add_argument('--input', type=str,
        required=True,
        help='input file name')
    parser.add_argument('--config', type=str,
        required=True,
        help='configuration file')
    parser.add_argument('--output', type=str,
        required=True,
        help='output file name')
    parser.add_argument('--logfile', type=str, 
        default=sys.stdout,
        help='log file (default stdout)')
    parser.add_argument('--debug', action='store_true',
        help='set logging level to debug')
    parser.add_argument('--create_time_dim', action='store_true',
        help='create time dimension in netcdf output')
    parser.add_argument('--ref_time_from_filename', action='store_true',
        help='extract forecast reference time from filename')
    args = parser.parse_args()

    """
    Setup logging
    """
    logging_level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(stream=args.logfile, level=logging_level)

    if os.path.isdir(args.datadir):
        logging.info(args.datadir)
    else:
        logging.error(args.datadir + ' not found')
        sys.exit(1)
    logging.info(args.input)
    logging.info(args.config)
    logging.info(args.output)

    """
    Construct input and output filenames
    """
    filename_in = os.path.join(args.datadir, args.input)
    filename_out = os.path.join(args.datadir, args.output)

    """
    Read YAML configuration file
    """
    config = yaml.load(
        open(args.config), Loader=yaml.FullLoader)
    logging.info(config)

    """
    Read dataset
    """
    try:
        if filename_in.split('.')[-1] == 'grb2' or 'pgrb2' in filename_in:
            logging.info('Opening GRIB2 ' + filename_in)
            ds = xr.open_dataset(filename_in, engine='cfgrib',
                backend_kwargs={'filter_by_keys': {'typeOfLevel': 'isobaricInhPa'}})
        else:
            logging.info('Opening NetCDF ' + filename_in)
            ds = xr.open_dataset(filename_in)
        logging.debug(ds)
    except:
        ds = xr.Dataset()
        logging.error('Unable to open ' + filename_in)
        logging.error(sys.exc_info()[0])

    """
    if 'valid_time' in ds:
        logging.info(datetime.utcfromtimestamp(
                     ds['valid_time'].astype('O')/1e9))
        logging.info(datetime.utcfromtimestamp(
                     ds['time'].astype('O')/1e9))
    """

    """
    Convert pressure levels to height levels
    """
    if (config['height_from_pressure']
        and config['vertical_coord_type_in'] == 'pressure'
        and config['vertical_coord_type_out'] == 'height'):

        surface_geopotential, surface_pressure, \
            temperature, relative_humidity \
            = read_required_fields(config, ds)

        layer_height = height_from_pressure(config,
            surface_geopotential, surface_pressure,
            temperature, relative_humidity)

    else:

        layer_height = ds[config['geopotential_height_name']]

    """
    Interpolate
    """
    ds_out = xr.Dataset()
    if 'valid_time' in ds:
        ds_out['valid_time'] = ds['valid_time'].values
        if 'time' in ds:
            ds_out['init_time'] = ds['time'].values

    ds_out['lat'] = ds['lat']
    ds_out['lon'] = ds['lon']

    for attr in ds.attrs:
        ds_out.attrs[attr] = ds.attrs[attr]

    for fieldname in config['fields']:

        field_interp, coords_interp = vertical_interp(fieldname, config,
            layer_height, ds[fieldname])

        ds_out[fieldname] = field_interp

    """
    Write dataset
    """
    try:
        logging.info('Creating with NetCDF4 ' + filename_out)
        ds_nc = nc.Dataset(filename_out, 'w')
        if args.ref_time_from_filename: 
            ref_time = filename_out.split('.')[1]
            write_dataset(ds_out, ds_nc, coords_interp=coords_interp,
                forecast_reference_time=ref_time, create_time_dim=args.create_time_dim)
        else:
            write_dataset(ds_out, ds_nc, coords_interp=coords_interp,
                create_time_dim=args.create_time_dim)
        ds_nc.close()
    except:
        logging.info('Creating with xarray ' + filename_out)
        ds_nc.to_netcdf(filename_out)

    logging.debug(os.system('ncdump -h ' + filename_out))



import numpy as np
import pytest

from metcalcpy.agg_eclv import AggEclv, pd


def test_calculate_value_and_ci(settings):
    agg_eclv = settings['agg_stat']
    agg_eclv.calculate_stats_and_ci()
    result_frame = pd.read_csv(
        agg_eclv.params['agg_stat_output'],
        header=[0],
        sep='\t'
    )
    assert result_frame.size == 304
    assert result_frame.shape == (38, 8)
    assert np.allclose(result_frame['y_pnt_i'][2], 0.5285295)
    assert np.allclose(result_frame['y_pnt_i'][26], 0.65747)
    assert np.allclose(result_frame['nstats'][0], 23)
    assert result_frame['stat_btcl'][9] <= result_frame['y_pnt_i'][9] <= result_frame['stat_btcu'][9]
    assert result_frame['stat_btcl'][24] <= result_frame['y_pnt_i'][24] <= result_frame['stat_btcu'][24]


@pytest.fixture
def settings():
    """Initialise values for testing.

    Returns:
        dictionary with values of different type
    """
    params = {'random_seed': 1, 'indy_var': '',
              'method': 'perc',
              'num_iterations': 100, 'event_equal': 'False',
              'agg_stat_input': 'data/agg_eclv_data.data',
              'agg_stat_output': 'data/agg_eclv_data_output.data',
              'fixed_vars_vals_input': {},
              'series_val_1': {'model': ['WRF'], 'fcst_lev': ['Z10', 'P850-700']},
              'alpha': 0.05, 'line_type': 'ctc',
              'num_threads': -1,
              'indy_vals': [],
              'agg_stat1': ['ECLV'],
              'circular_block_bootstrap': True,
              'equalize_by_indep': 'True',
              'cl_step': 0.05
              }
    agg_stat = AggEclv(params)
    settings_dict = dict()
    settings_dict['agg_stat'] = agg_stat
    return settings_dict



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: agg_eclv.py

How to use:
 - Call from other Python function
        AGG_STAT = AggEclv(PARAMS)
        AGG_STAT.calculate_stats_and_ci()
        where PARAMS – a dictionary with data description parameters including
        location of input and output data.
        The structure is similar to Rscript template

 - Run as a stand-alone script
        python agg_eclv.py <parameters_file>
        where - <parameters_file> is YAML file with parameters
        and environment variable should be set to PYTHONPATH=<path_to_METcalcpy>

 - Run from Java
        proc = Runtime.getRuntime().exec(
                “python agg_eclv.py <parameters_file>”,
                new String[]{”PYTHONPATH=<path_to_METcalcpy>”},
                new File(System.getProperty("user.home")));

"""
import math
import sys
import itertools
import argparse
import yaml
import pandas as pd

from metcalcpy.bootstrap import bootstrap_and_value, BootstrapResults
from metcalcpy.event_equalize import event_equalize
from metcalcpy.util.utils import PRECISION, is_string_strictly_float
from metcalcpy.util.eclv_statistics import *

from metcalcpy.util.utils import is_string_integer, parse_bool

__author__ = 'Tatiana Burek'


class AggEclv:
    """A class that performs aggregation statistic logic for ECLV data type on input data frame.
           All parameters including data description and location is in the parameters dictionary
           Usage:
                initialise this call with the parameters dictionary and than
                call calculate_stats_and_ci method
                This method will crate and save to the file aggregation statistics
                    agg_stat = AggEclv(params)
                    agg_stat.calculate_stats_and_ci()
            Raises: EmptyDataError or ValueError when the input DataFrame is empty
                or doesn't have data
       """
    LINE_TYPE_COLUMNS = {
        'ctc': ["total", "fy_oy", "fy_on", "fn_oy", "fn_on"],
        'pct': ["thresh_i", "oy_i", "on_i"]
    }
    HEADER = ['thresh_i', 'x_pnt_i', 'y_pnt_i', 'stat_btcl', 'stat_btcu', 'nstats']

    def __init__(self, in_params):
        """Initialises the class by saving input parameters and reading data from file

            Args:
                in_params - input parameters as a dictionary
            Raises: EmptyDataError or ValueError when the input DataFrame is empty
                or doesn't have data
        """

        self.statistic = None
        self.current_thresh = None
        self.params = in_params
        self.steps = np.arange(self.params['cl_step'], 1, self.params['cl_step'])
        self.column_names = np.array(self.LINE_TYPE_COLUMNS[self.params['line_type']])

        if 'add_base_rate' in self.params.keys():
            self.add_base_rate = self.params['add_base_rate']
        else:
            self.add_base_rate = 0
        if self.add_base_rate != 0 or self.add_base_rate != 1:
            self.add_base_rate = 0

        try:
            self.input_data = pd.read_csv(
                self.params['agg_stat_input'],
                header=[0],
                sep='\t'
            )
        except pd.errors.EmptyDataError:
            raise
        except KeyError as er:
            print(f'ERROR: parameter with key {er} is missing')
            raise
        self.group_to_value = {}

    def _calc_stats(self, values):
        """Calculate the statistic of values for each bootstrap sample
            Args:
                values: a np.array of values we want to calculate the statistic on
                    This is actually a 2d array (matrix) of values. Each row represents
                    a bootstrap resample simulation that we wish to aggregate across.
             Returns:
                a list of calculated statistics
            Raises:
                an error

        """

        if values is not None and values.ndim == 2:
            # The single value case
            stat_values = [
                calculate_eclv(values, self.column_names, self.current_thresh, self.params['line_type'], self.steps,
                               self.add_base_rate)]

        elif values is not None and values.ndim == 3:
            # bootstrapped case
            stat_values = []
            for row in values:
                stat_value = [
                    calculate_eclv(row, self.column_names, self.current_thresh, self.params['line_type'], self.steps)]
                stat_values.append(stat_value)

        else:
            raise KeyError("can't calculate statistic")
        return stat_values

    def _get_bootstrapped_stats(self, series_data, thresholds):
        """ Calculates aggregation statistic value and CI intervals if needed for input data
            Args:
                series_data: pandas data frame
            Returns:
                BootstrapDistributionResults object

        """

        # if the data frame is empty - do nothing and return an empty object
        if series_data.empty:
            return BootstrapResults(lower_bound=None,
                                    value=None,
                                    upper_bound=None)

        data = series_data[self.column_names].to_numpy()
        boot_stat_thresh = {}
        for ind, thresh in enumerate(thresholds):
            self.current_thresh = thresh
            if self.params['num_iterations'] == 1:
                # don't need bootstrapping and CI calculation -
                # calculate the statistic and exit
                stat_val = self._calc_stats(data)[0]

                results = BootstrapResults(lower_bound=None,
                                           value=stat_val,
                                           upper_bound=None)

            else:
                # need bootstrapping and CI calculation in addition to statistic
                try:
                    block_length = 1
                    # to use circular block bootstrap or not
                    is_cbb = True
                    if 'circular_block_bootstrap' in self.params.keys():
                        is_cbb = parse_bool(self.params['circular_block_bootstrap'])

                    if is_cbb:
                        block_length = int(math.sqrt(len(data)))
                    results = bootstrap_and_value(
                        data,
                        stat_func=self._calc_stats,
                        num_iterations=self.params['num_iterations'],
                        num_threads=self.params['num_threads'],
                        ci_method=self.params['method'],
                        save_data=False,
                        block_length=block_length,
                        eclv=True
                    )

                except KeyError as err:
                    results = BootstrapResults(None, None, None)
                    print(err)
            boot_stat_thresh[ind] = results

        return boot_stat_thresh

    def _init_out_frame(self, fields, row_number):
        """ Initialises the output frame and add series values to each row
            Args:
                series_fields: list of all possible series fields
                series: list of all series definitions
            Returns:
                pandas data frame
        """
        result = pd.DataFrame()
        # fill series variables and values
        for field in fields:
            if field == 'nstats':
                result[field] = [0] * row_number
            else:
                result[field] = [None] * row_number
        return result

    def _proceed_with_axis(self):
        """Calculates stat values for the requested Y axis

             Returns:
                pandas dataframe  with calculated stat values and CI

        """
        if self.input_data.empty:
            return pd.DataFrame()

        series_val = self.params['series_val_1']
        if len(series_val) > 0:
            current_header = list(series_val.keys())
            current_header.extend(self.HEADER)
        else:
            current_header = self.HEADER.copy()

        all_points = list(itertools.product(*series_val.values()))

        out_frame = self._init_out_frame(current_header, 0)

        # for each point
        for point in all_points:
            out_frame_local = self._init_out_frame(current_header, len(self.steps) + self.add_base_rate)
            # filter point data
            all_filters = []
            for field_ind, field in enumerate(series_val.keys()):
                filter_value = point[field_ind]
                filter_list = [filter_value]
                out_frame_local[field] = [filter_value] * (len(self.steps))
                for i, filter_val in enumerate(filter_list):
                    if is_string_integer(filter_val):
                        filter_list[i] = int(filter_val)
                    elif is_string_strictly_float(filter_val):
                        filter_list[i] = float(filter_val)
                if field in self.input_data.keys():
                    all_filters.append((self.input_data[field].isin(filter_list)))

            # use numpy to select the rows where any record evaluates to True
            mask = np.array(all_filters).all(axis=0)
            point_data = self.input_data.loc[mask]

            # calculate bootstrap results
            if 'thresh_i' in point_data.columns:
                thresholds = point_data['thresh_i'].unique().tolist()
                thresholds.sort()
            else:
                thresholds = [0]

            bootstrap_results = self._get_bootstrapped_stats(point_data, thresholds)

            for thresh_ind, thresh in enumerate(thresholds):
                out_frame_local['thresh_i'] = [thresh] * (len(self.steps) + self.add_base_rate)
                out_frame_local['x_pnt_i'] = bootstrap_results[thresh_ind].value['cl']
                out_frame_local['y_pnt_i'] = bootstrap_results[thresh_ind].value['V']
                out_frame_local['nstats'] = [len(point_data)] * (len(self.steps) + self.add_base_rate)
                if self.params['num_iterations'] > 1:
                    out_frame_local['stat_btcu'] = bootstrap_results[thresh_ind].upper_bound
                    out_frame_local['stat_btcl'] = bootstrap_results[thresh_ind].lower_bound
                frames = [out_frame, out_frame_local]
                out_frame = pd.concat(frames)

        out_frame.reset_index(inplace=True, drop=True)
        return out_frame

    def calculate_stats_and_ci(self):
        """ Calculates aggregated statistics and confidants intervals
            ( if parameter num_iterations > 1) for each series point
            Writes output data to the file

        """

        # set random seed if present
        if self.params['random_seed'] is not None and self.params['random_seed'] != 'None':
            np.random.seed(self.params['random_seed'])

        is_event_equal = parse_bool(self.params['event_equal'])
        # perform EE if needed
        if is_event_equal:
            fix_vals_permuted_list = []

            for key in self.params['fixed_vars_vals_input']:
                vals_permuted = list(itertools.product(*self.params['fixed_vars_vals_input'][key].values()))
                vals_permuted_list = [item for sublist in vals_permuted for item in sublist]
                fix_vals_permuted_list.append(vals_permuted_list)

            fix_vals_keys = list(self.params['fixed_vars_vals_input'].keys())
            is_equalize_by_indep = parse_bool(self.params['equalize_by_indep'])
            self.input_data = event_equalize(self.input_data, 'stat_name',
                                             self.params['series_val_1'],
                                             fix_vals_keys,
                                             fix_vals_permuted_list, is_equalize_by_indep, False)

        out_frame = self._proceed_with_axis()
        header = True
        mode = 'w'

        if 'append_to_file' in self.params.keys() and self.params['append_to_file'] == 'True':
            header = False
            mode = 'a'

        export_csv = out_frame.to_csv(self.params['agg_stat_output'],
                                      index=None, header=header, mode=mode,
                                      sep="\t", na_rep="NA", float_format='%.' + str(PRECISION) + 'f')


if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(description='List of agg_stat arguments')
    PARSER.add_argument("parameters_file", help="Path to YAML parameters file",
                        type=argparse.FileType('r'),
                        default=sys.stdin)
    ARGS = PARSER.parse_args()
    PARAMS = yaml.load(ARGS.parameters_file, Loader=yaml.FullLoader)

    AGG_STAT = AggEclv(PARAMS)
    AGG_STAT.calculate_stats_and_ci()



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
This is an example script for cross-spectral analysis. Computes power and cross-spectra for multiple input
data sets.
"""
import argparse
import xarray as xr
import numpy as np

"""
local scripts, if loading from a different directory include that with a '.' between
directory name and script name
"""
from spacetime import mjo_cross
from spacetime import get_symmasymm
from spacetime_utils import save_Spectra

"""
Set parameters for the spectra calculation.
spd:  Number of observations per day.
nperseg:  Number of data points per segment.
segOverlap:  How many data points of overlap between segments. If negative there is overlap
             If positive skip that number of values between segments.
Symmetry:    Can be "symm", "asymm" for symmetric or anti-symmetric across the equator. "latband" 
             if the spectra averaged across the entire latitude band are requested.
latMin:      Minimum latitude to include.
latMax:      Maximum latitude to include.
"""
spd = 2
nperseg = 46 * spd
segOverLap = -20 * spd
Symmetry = "symm"
latMin = -15.
latMax = 15.
datestrt = '2015-12-01'  # plot start date, format: yyyy-mm-dd
datelast = '2016-03-31'  # plot end date, format: yyyy-mm-dd

parser = argparse.ArgumentParser()
parser.add_argument('--datapath', type=str,
    help='data path')
parser.add_argument('--pathout', type=str,
    help='output path')
args = parser.parse_args()

# datapath = '../data/'
# path to the location where to save the output spectra
# pathout = '../data/'
datapath = args.datapath
pathout = args.pathout

print("reading data from file:")
""" 
Read in data here. Example:
"""
ds = xr.open_dataset(datapath+'precip.trmm.'+str(spd)+'x.1p0.v7a.fillmiss.comp.2014-2016.nc')
z = ds.precip
z = z.sel(lat=slice(latMin, latMax))
z = z.sel(time=slice(datestrt, datelast))
z = z.squeeze()
latC = ds.lat.sel(lat=slice(latMin, latMax))

ds = xr.open_dataset(datapath+'precip.erai.sfc.1p0.'+str(spd)+'x.2014-2016.nc')
x = ds.precip
x = x.sel(lat=slice(latMin, latMax))
x = x.sel(time=slice(datestrt, datelast))
x = x.squeeze()
latA = ds.lat.sel(lat=slice(latMin, latMax))
ds.close()

ds = xr.open_dataset(datapath+'div.erai.850.1p0.'+str(spd)+'x.2014-2016.nc')
y = ds.div
y = y.sel(lat=slice(latMin, latMax))
y = y.sel(time=slice(datestrt, datelast))
y = y.squeeze()
latB = ds.lat.sel(lat=slice(latMin, latMax))

ds = xr.open_dataset(datapath+'div.erai.200.1p0.'+str(spd)+'x.2014-2016.nc')
w = ds.div
w = y.sel(lat=slice(latMin, latMax))
w = y.sel(time=slice(datestrt, datelast))
w = y.squeeze()
latB = ds.lat.sel(lat=slice(latMin, latMax))



if any(latA - latB) != 0:
    print("Latitudes must be the same for both variables! Check latitude ordering.")

print("get symmetric/anti-symmetric components:")
if Symmetry == "symm" or Symmetry == "asymm":
    X = get_symmasymm(x, latA, Symmetry)
    Y = get_symmasymm(y, latB, Symmetry)
    Z = get_symmasymm(z, latC, Symmetry)
    W = get_symmasymm(w, latB, Symmetry)
else:
    X = x
    Y = y
    Z = z
    W = w

print("compute cross-spectrum:")
"""
The output from mjo_cross includes:
STC = [8,nfreq,nwave]
STC[0,:,:] : Power spectrum of x
STC[1,:,:] : Power spectrum of y
STC[2,:,:] : Co-spectrum of x and y
STC[3,:,:] : Quadrature-spectrum of x and y
STC[4,:,:] : Coherence-squared spectrum of x and y
STC[5,:,:] : Phase spectrum of x and y
STC[6,:,:] : Phase angle v1
STC[7,:,:] : Phase angle v2
freq 
wave
number_of_segments
dof
prob
prob_coh2
"""
result = mjo_cross(X, Y, nperseg, segOverLap)
STC = result['STC']  # , freq, wave, number_of_segments, dof, prob, prob_coh2
freq = result['freq']
freq = freq * spd
wnum = result['wave']
# save spectra in netcdf file
fileout = 'SpaceTimeSpectra_ERAI_P_D850_' + Symmetry + '_' + str(spd) + 'spd'
print('saving spectra to file: ' + pathout + fileout + '.nc')
save_Spectra(STC, freq, wnum, fileout, pathout)

result = mjo_cross(X, W, nperseg, segOverLap)
STC = result['STC']  # , freq, wave, number_of_segments, dof, prob, prob_coh2
freq = result['freq']
freq = freq * spd
wnum = result['wave']
# save spectra in netcdf file
fileout = 'SpaceTimeSpectra_ERAI_P_D200_' + Symmetry + '_' + str(spd) + 'spd'
print('saving spectra to file: ' + pathout + fileout + '.nc')
save_Spectra(STC, freq, wnum, fileout, pathout)

result = mjo_cross(X, Z, nperseg, segOverLap)
STC = result['STC']  # , freq, wave, number_of_segments, dof, prob, prob_coh2
freq = result['freq']
freq = freq * spd
wnum = result['wave']
# save spectra in netcdf file
fileout = 'SpaceTimeSpectra_ERAI_TRMM_P_' + Symmetry + '_' + str(spd) + 'spd'
print('saving spectra to file: ' + pathout + fileout + '.nc')
save_Spectra(STC, freq, wnum, fileout, pathout)



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** CIRES, Regents of the University of Colorado
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
import numpy as np
import xarray as xr


def zonal_mean(dat,dimvar='longitude'):
    """Compute the zonal mean.
    Parameters
    ----------
    dat : `xarray.DataArray` or `xarray.Dataset`
        data containing a dimension that you want to compute a zonal mean on
    dimvar: Name of the dimension to compute the zonal mean.  Longitude is the 
        default if it's not specified
    Returns
    -------
    `xarray.DataArray` or `xarray.Dataset`
        the mean across the zonal dimension
    """
    return dat.mean(dimvar)


def meridional_mean(dat, lat1, lat2, dimvar='latitude'):
    """Compute the cos(lat) weighted mean of a quantity between two latitudes.
    Parameters
    ----------
    dat : `xarray.DataArray` or `xarray.Dataset`
        data containing a dimension that you want ot compute a meridional mean
        on that spans lat1 and lat2
    lat1 : float
        The beginning latitude limit of the band average.  This should always be less 
        than lat2
    lat2 : float
        The ending latitude limit of the band average.  This should always be greater
        than lat1
    dimvar: Name of the dimension to compute the meridional mean.  Latitude is the
        default if it's not specified
    Returns
    -------
    `xarray.DataArray` or `xarray.Dataset`
        the weighted mean across the latitude dimension limited
        by lat1 and lat2
    """

    # Check inputs
    if lat1 > lat2:
        raise ValueError('lat1 is greater than lat2, but it must be less than lat2')
    elif lat1 == lat2:
        raise ValueError('lat1 is equal to lat2, but it must be less than lat2')

    wgts = np.cos(np.deg2rad(dat[dimvar].where((dat[dimvar] >= 60) & (dat[dimvar] <= 90),drop=True)))

    return dat.where((dat[dimvar] >= 60) & (dat[dimvar] <= 90),drop=True).weighted(wgts).mean(dimvar)



# ============================*
 # ** Copyright UCAR (c) 2020
 # ** University Corporation for Atmospheric Research (UCAR)
 # ** National Center for Atmospheric Research (NCAR)
 # ** Research Applications Lab (RAL)
 # ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
 # ============================*
 
 
 
"""
Program Name: event_equalize.py
"""

__author__ = 'Tatiana Burek'

import itertools
from typing import Any
import re
import numpy as np
import pandas as pd

from metcalcpy import GROUP_SEPARATOR, DATE_TIME_REGEX


def event_equalize(series_data, indy_var, series_var_vals, fix_vars,
                   fix_vals_permuted, equalize_by_indep, multi):
    """Performs event equalisation.

    event_equalize assumes that the input series_data contains data indexed by fcst_valid_beg,
    series values and the independent variable values.  It builds a new dataframe which
    contains the same data except for records that don't have corresponding fcst_valid_beg
    values in each other series record with the same independent variable value.
    For example, if the series_a has data valid at 20111010_000000 for F12 and series_b does not,
    the series_a record is removed.

    Args:
        series_data: data frame containing the records to equalize, including fcst_valid_beg, series
                values and independent variable values
        indy_var: name of the independent variable
        series_var_vals: series variable names and values
        fix_vars: name of the fixed variable
        fix_vals_permuted:  fixed variable values to equalize over
        equalize_by_indep:  include or not the independent variable to EE
        multi: FALSE for normal event equalization, TRUE for equalization of mulitple events
                at each combination of fcst_valid_beg, series values and independent value
                - for example with MODE objects
    Returns:
        A data frame that contains equalized records
    """
    pd.options.mode.chained_assignment = None
    column_names = list(series_data)
    exception_columns = ["", "fcst_valid_beg", 'fcst_lead', 'fcst_valid', 'fcst_init', 'fcst_init_beg']
    if isinstance(fix_vars, str):
        fix_vars = [fix_vars]
    if 'fcst_valid_beg' in column_names:
        series_data['equalize'] = series_data.loc[:, 'fcst_valid_beg'].astype(str) \
                                  + ' ' + series_data.loc[:, 'fcst_lead'].astype(str)
    elif 'fcst_valid' in column_names:
        series_data['equalize'] = series_data.loc[:, 'fcst_valid'].astype(str) + ' ' \
                                  + series_data.loc[:, 'fcst_lead'].astype(str)

    # add independent variable if needed
    if equalize_by_indep and indy_var not in exception_columns:
        series_data['equalize'] = series_data.loc[:, 'equalize'].astype(str) + " " \
                                  + series_data.loc[:, indy_var].astype(str)

    vars_for_ee = dict()

    # add series variables
    if series_var_vals:
        for series_var, series_vals in series_var_vals.items():
            if series_var not in exception_columns:
                series_vals_no_groups = []
                if isinstance(series_vals, str):
                    series_vals = [series_vals]
                for series_val in series_vals:
                    actual_vals = re.findall(DATE_TIME_REGEX, series_val)
                    if len(actual_vals) == 0:
                        actual_vals = series_val.split(GROUP_SEPARATOR)
                    series_vals_no_groups.extend(actual_vals)
                vars_for_ee[series_var] = series_vals_no_groups

    # add fixed variables if present
    if fix_vars:
        for var_for_ee_ind, fix_var in enumerate(fix_vars):
            if fix_var not in exception_columns:
                vals = fix_vals_permuted[var_for_ee_ind]
                if isinstance(vals, str):
                    vals = [vals]
                vars_for_ee[fix_var] = vals

    # create a list of permutations representing the all variables for
    vars_for_ee_permuted = list(itertools.product(*vars_for_ee.values()))

    equalization_cases = []

    # for each permutation

    for permutation_index, permutation in enumerate(vars_for_ee_permuted):

        permutation_data = series_data.copy()
        for var_for_ee_ind, var_for_ee in enumerate(list(vars_for_ee)):
            if is_string_integer(permutation[var_for_ee_ind]):
                permutation_data = permutation_data[permutation_data[var_for_ee]
                                                    == int(permutation[var_for_ee_ind])]
            else:
                if permutation[var_for_ee_ind] == 'NA':
                    permutation_data = permutation_data[permutation_data[var_for_ee].isnull()]
                else:
                    permutation_data = permutation_data[permutation_data[var_for_ee]
                                                        == permutation[var_for_ee_ind]]
        # if the list contains repetitive values, show a warning
        if multi is False and len(permutation_data['equalize']) \
                != len(set(permutation_data['equalize'])):
            print(
                f"WARNING: eventEqualize() detected non-unique events for {permutation}"
                f" using [fcst_valid_beg,fcst_lead)]")

        if permutation_index == 0:
            # init the equalization list
            equalization_cases = permutation_data['equalize'].reset_index(drop=True)
        else:
            # if there is an equalization list, equalize the current series data

            # find indexes of common cases that are on the current cases list
            common_cases_ind = \
                pd.Series(equalization_cases, dtype=str).isin(set(permutation_data['equalize']))

            # identify discarded cases for this permutation
            discarded_cases: list = equalization_cases[~common_cases_ind]

            # add cases that are in current permutation
            # but not in the common cases and add them to the discarded list
            permutation_cases_not_in_common_cases_ind = \
                pd.Series(permutation_data['equalize'], dtype=str).isin(equalization_cases)
            permutation_cases_not_in_common_cases: Any = \
                permutation_data['equalize'][~permutation_cases_not_in_common_cases_ind]

            discarded_cases = pd.concat([discarded_cases, permutation_cases_not_in_common_cases])
            # report the discarded records
            for discarded_case in discarded_cases:
                print(f"WARNING: discarding series member with case {discarded_case}"
                      f" for {permutation}")

            # update the equalization list by removing records
            equalization_cases = equalization_cases[common_cases_ind]
            equalization_cases.reset_index(drop=True)

    # remove data with discarded cases from the main frame
    equalization_cases_ind = \
        pd.Series(series_data['equalize'], dtype=str).isin(equalization_cases)
    series_data_ee = series_data[equalization_cases_ind]

    if len(series_data_ee) != len(series_data):
        print(f"WARNING: event equalization removed {len(series_data) - len(series_data_ee)} rows")

    return series_data_ee


def is_string_integer(str_int):
    """Checks if the input string is integer.

         Args:
             str_int: string value to check

        Returns:
            True - if the input value is an integer
            False - if the input value is not an integer
    """
    try:
        int(str_int)
        return True
    except (ValueError, TypeError):
        return False



"""Tests the operation of METcalcpy's statistics code."""
import numpy as np
import pytest
from metcalcpy.util.met_stats import get_column_index_by_name
from metcalcpy.util.correlation import corr, acf
from metcalcpy.util.utils import round_half_up
from metcalcpy.util.wald_wolfowitz_runs_test import runs_test
from metcalcpy.util.eclv_statistics import calculate_eclv


def test_get_column_index_by_name(settings):
    column_name = 'fobar'
    assert 11 == get_column_index_by_name(settings['columns'], column_name)

    column_name = 'not_in_array'
    assert not get_column_index_by_name(settings['columns'], column_name)


@pytest.fixture
def settings():
    """Initialise values for testing.

    Returns:
        dictionary with values of different type
    """
    settings_dict = dict()
    columns = np.array(['model', 'fcst_init_beg', 'fcst_valid_beg', 'fcst_lead', 'vx_mask', 'fcst_var',
                        'stat_name', 'stat_value', 'total', 'fbar', 'obar', 'fobar', 'ffbar', 'oobar',
                        'mae'])
    settings_dict['columns'] = columns

    return settings_dict


def test_corr():
    x = [103.4, 59.92, 68.17, 94.54, 69.48, 72.17, 74.37, 84.44, 96.74, 94.26, 48.52, 95.68]
    y = [90.11, 77.71, 77.71, 97.51, 58.21, 101.3, 79.84, 96.06, 89.3, 97.22, 61.62, 85.8]

    corr_val = corr(x=x, y=y)['r'].tolist()[0]
    assert round_half_up(corr_val, 2) == 0.67


def test_acf():
    x = [2.4, 2.4, 2.4, 2.2, 2.1, 1.5, 2.3, 2.3, 2.5, 2.0, 1.9, 1.7, 2.2, 1.8, 3.2, 3.2, 2.7, 2.2, 2.2, 1.9, 1.9, 1.8,
         2.7, 3.0, 2.3, 2.0, 2.0, 2.9, 2.9, 2.7, 2.7, 2.3, 2.6, 2.4, 1.8, 1.7, 1.5, 1.4, 2.1, 3.3, 3.5, 3.5, 3.1, 2.6,
         2.1, 3.4, 3.0, 2.9]
    acf_val = acf(x, 'correlation', lag_max=10)
    assert -0.154 == round_half_up(acf_val[10], 3)
    acf_val = acf(x, 'correlation')
    assert 0.151 == round_half_up(acf_val[16], 3)


def test_eclv():
    x = np.array([[666, 112, 25, 33, 496], [350, 73, 9, 22, 246], [316, 39, 16, 11, 250]])
    columns_names = np.array(['total', 'fy_oy', 'fy_on', 'fn_oy', 'fn_on'])
    cl_pts = [0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95]
    thresh = 0
    line_type = 'ctc'
    add_base_rate = 1
    eclv = calculate_eclv(x, columns_names, thresh, line_type, cl_pts, add_base_rate)
    assert 0.7244291 == eclv['vmax']
    assert 0.0479846 == eclv['F']
    assert 0.7724138 == eclv['H']
    assert 0.2177177 == eclv['s']
    assert 20 == len(eclv['V'])
    assert -0.2514395393474087 == eclv['V'][0]



def test_runs_test():
    x = [2.4, 2.4, 2.4, 2.2, 2.1, 1.5, 2.3, 2.3, 2.5, 2.0, 1.9, 1.7, 2.2, 1.8, 3.2, 3.2, 2.7, 2.2, 2.2, 1.9, 1.9, 1.8,
         2.7, 3.0, 2.3, 2.0, 2.0, 2.9, 2.9, 2.7, 2.7, 2.3, 2.6, 2.4, 1.8, 1.7, 1.5, 1.4, 2.1, 3.3, 3.5, 3.5, 3.1, 2.6,
         2.1, 3.4, 3.0, 2.9]

    ww_run = runs_test(x, 'left.sided', 'median')
    assert 0.00117 == round_half_up(ww_run['p_value'], 5)


if __name__ == "__main__":
    test_get_column_index_by_name()
    test_corr()
    test_acf()
    test_runs_test()
    test_eclv()



import pytest
import numpy as np
import metcalcpy.util.utils as utils

def test_lon_from_360_to_180():

    # Verify that longitude type is maintained, ie an int isn't converted to
    # a float when converting coord values.
    i_lon = [i_lon for i_lon in range (0, 359)]
    np_lon = np.linspace(0, 359, 360)

    i_west_east = utils.convert_lon_360_to_180(i_lon)
    np_west_east = utils.convert_lon_360_to_180(np_lon)


    assert isinstance(i_west_east[0], np.int64)
    assert isinstance(np_west_east[0], np.float)


    # Verify that some values were correctly converted
    # and that the ordering in the array is from negative to positive
    assert np_west_east[0] == -180.0
    assert np_west_east[359] == 179.0


if __name__ == "__main__":
    test_lon_from_360_to_180()







import numpy as np
import pytest
import math
import statistics

from metcalcpy import bootstrap
from metcalcpy.agg_stat import AggStat, pd
from metcalcpy.bootstrap import bootstrap_and_value
from metcalcpy.util.utils import round_half_up, PRECISION

TEST_LENGTH = 1000


def lossdiff_ml(data, ):
    if len(data.shape) < 3:
        lossdiff = data[:, 0] - data[:, 1]
        return [statistics.mean(lossdiff)]
    else:
        result = []
        for i in range(0, data.shape[0]):
            lossdiff = data[i][:, 0] - data[i][:, 1]
            result.append(statistics.mean(lossdiff))
        return result


def lossdiff_mal(data):
    if len(data.shape) < 3:
        ALlossdiff = abs(data[:, 0]) - abs(data[:, 1])
        return [statistics.mean(ALlossdiff)]
    else:
        result = []
        for i in range(0, data.shape[0]):
            ALlossdiff = abs(data[i][:, 0]) - abs(data[i][:, 1])
            result.append(statistics.mean(ALlossdiff))
        return result


def lossdiff_msl(data):
    if len(data.shape) < 3:
        SLlossdiff = data[:, 0] * data[:, 0] - data[:, 1] * data[:, 1]
        return [statistics.mean(SLlossdiff)]
    else:
        result = []
        for i in range(0, data.shape[0]):
            SLlossdiff = data[i][:, 0] * data[i][:, 0] - data[i][:, 1] * data[i][:, 1]
            result.append(statistics.mean(SLlossdiff))
        return result

@pytest.mark.skip("Not to be run in regression testing, due to extensive "
                  "number of data points.  This test takes a long time to run.")
def test_cboot():

    et = np.loadtxt(
        "./data/et.txt" )

    # create an array for accepted/rejected flags
    ml_reject = [1] * TEST_LENGTH
    mal_reject = [1] * TEST_LENGTH
    msl_reject = [1] * TEST_LENGTH
    mean_reject = [1] * TEST_LENGTH
    # run the boot ci TEST_LENGTH times
    for ind in range(TEST_LENGTH):
        results_ml = bootstrap_and_value(
            et,
            stat_func=lossdiff_ml,
            num_iterations=500, alpha=0.05,
            num_threads=1, ci_method='perc', block_length=32)

        # record if 0 in ci bounds (accept) = 0 or not (reject) = 1
        if results_ml.lower_bound <= 0 and results_ml.upper_bound >= 0:
            ml_reject[ind] = 0

        results_mal = bootstrap_and_value(
            et,
            stat_func=lossdiff_mal,
            num_iterations=500, alpha=0.05,
            num_threads=1, ci_method='perc', block_length=32)
        # record if 0 in ci bounds (accept)=0 or not (reject)=1
        if results_mal.lower_bound <= 0 and results_mal.upper_bound >= 0:
            mal_reject[ind] = 0

        results_msl = bootstrap_and_value(
            et,
            stat_func=lossdiff_msl,
            num_iterations=500, alpha=0.05,
            num_threads=1, ci_method='perc', block_length=32)

        # record if 0 in ci bounds (accept)=0 or not (reject)=1
        if results_msl.lower_bound <= 0 and results_msl.upper_bound >= 0:
            msl_reject[ind] = 0

        results_mean = bootstrap_and_value(
            et[:, 0],
            stat_func=bootstrap.mean,
            num_iterations=500, alpha=0.05,
            num_threads=1, ci_method='perc', block_length=32)
        if results_mean.lower_bound <= 0 and results_mean.upper_bound >= 0:
            mean_reject[ind] = 0

    # get the number of rejected
    ml_number_of_rejected = sum(x == 1 for x in ml_reject)
    ml_percent_of_rejected = ml_number_of_rejected * 100 / TEST_LENGTH
    ml_frequencies_of_ml_rejected = ml_number_of_rejected / TEST_LENGTH

    msl_number_of_rejected = sum(x == 1 for x in msl_reject)
    msl_percent_of_rejected = msl_number_of_rejected * 100 / TEST_LENGTH
    msl_frequencies_of_rejected = msl_number_of_rejected / TEST_LENGTH

    mal_number_of_rejected = sum(x == 1 for x in mal_reject)
    mal_percent_of_rejected = mal_number_of_rejected * 100 / TEST_LENGTH
    mal_frequencies_of_rejected = mal_number_of_rejected / TEST_LENGTH

    mean_number_of_rejected = sum(x == 1 for x in mean_reject)
    mean_percent_of_rejected = mean_number_of_rejected * 100 / TEST_LENGTH
    mean_frequencies_of_rejected = mean_number_of_rejected / TEST_LENGTH

    print('for ML   p = {} total rejected = {} frequency = {}'.format(ml_percent_of_rejected, ml_percent_of_rejected,
                                                                      ml_frequencies_of_ml_rejected))
    print('for MAL  p = {} total rejected = {} frequency = {}'.format(mal_percent_of_rejected, mal_number_of_rejected,
                                                                      mal_frequencies_of_rejected))
    print('for MSL  p = {} total rejected = {} frequency = {}'.format(msl_percent_of_rejected, msl_number_of_rejected,
                                                                      msl_frequencies_of_rejected))
    print(
        'for mean  p = {} total rejected = {} frequency = {}'.format(mean_percent_of_rejected, mean_number_of_rejected,
                                                                     mean_frequencies_of_rejected))

@pytest.mark.skip("Not to be run in as a regression test, it uses an extensive number of points"
                  " and takes a long time (well beyond 5 minutes) to run.")
def test_boot():
    # size of array
    n = 100
    for mean in range(6):
        p = get_rejected(mean, n)
        print('for mean = {} p = {}'.format(mean, p))


def get_rejected(mean, n):
    """Calculate the percent of rejected values for 0 hypothesis test
        for CI for mean statistic of the normal distribution of 100 values

        Args:
            mean - mean value for the normal distribution
            n - size of array
        Returns:
            percent of rejected values
    """

    block_lenght = int(math.sqrt(n))

    # create an array for accepted/rejected flags
    reject = [1] * TEST_LENGTH
    # run the boot ci TEST_LENGTH times
    for ind in range(TEST_LENGTH):
        # create normal distribution
        data = np.random.normal(loc=mean, size=n, scale=10)
        # get ci for mean stat for this distribution
        results = bootstrap_and_value(
            data,
            stat_func=bootstrap.mean,
            num_iterations=500, alpha=0.05,
            num_threads=1, ci_method='perc', block_length=block_lenght)

        # record if 0 in ci bounds (accept) or not (reject)
        if results.lower_bound <= 0 and results.upper_bound >= 0:
            reject[ind] = 0

    # get the number of rejected
    number_of_rejected = sum(x == 1 for x in reject)
    percent_of_rejected = number_of_rejected * 100 / TEST_LENGTH
    return percent_of_rejected


def test_prepare_sl1l2_data(settings):
    agg_stat = settings['agg_stat']
    series_data = agg_stat.input_data[
        (agg_stat.input_data['model'] == 'ENS001v3.6.1_d01')
        & (agg_stat.input_data["fcst_lead"] == 60000)]
    agg_stat.statistic = agg_stat.params['list_stat_1'][0].lower()
    agg_stat._prepare_sl1l2_data(series_data)
    result = np.array([6532.28230, 4034.79153, 7108.67791, 5379.89956])
    assert np.allclose(result, series_data['fbar'])


def test_calc_stats(settings):
    agg_stat = settings['agg_stat']
    series_data = agg_stat.input_data[
        (agg_stat.input_data['model'] == 'ENS001v3.6.1_d01')
        & (agg_stat.input_data["fcst_lead"] == 60000)]
    agg_stat.statistic = agg_stat.params['list_stat_1'][0].lower()
    agg_stat._prepare_sl1l2_data(series_data)
    stat_val = agg_stat._calc_stats(series_data.values)
    assert np.allclose(np.array([192.13042749999997]), stat_val)


def test_calc_stats_derived(settings):
    agg_stat = settings['agg_stat']
    series_data_1 = agg_stat.input_data[
        (agg_stat.input_data['model'] == 'ENS001v3.6.1_d01')
        & (agg_stat.input_data["fcst_lead"] == 60000)]
    agg_stat.statistic = agg_stat.params['list_stat_1'][0].lower()
    agg_stat._prepare_sl1l2_data(series_data_1)

    series_data_2 = agg_stat.input_data[
        (agg_stat.input_data['model'] == 'ENS001v3.6.1_d02')
        & (agg_stat.input_data["fcst_lead"] == 60000)]
    agg_stat._prepare_sl1l2_data(series_data_2)

    operation = np.full((len(series_data_1.values), 1), "DIFF")
    values_both_arrays = np.concatenate((series_data_1.values, series_data_2.values), axis=1)
    values_both_arrays = np.concatenate((values_both_arrays, operation), axis=1)
    stat_val = agg_stat._calc_stats_derived(values_both_arrays)

    assert np.allclose(np.array([0.00362229]), stat_val)


def test_get_derived_series(settings):
    agg_stat = settings['agg_stat']
    series_val = agg_stat.params['series_val_1']
    indy_vals = agg_stat.params['indy_vals']
    result = agg_stat._get_derived_points(series_val, indy_vals)
    expected = [('DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)', '0', 'FBAR'),
                ('DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)', '30000', 'FBAR'),
                ('DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)', '60000', 'FBAR'),
                ('DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)', '90000', 'FBAR'),
                ('DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)', '120000', 'FBAR'),
                ('DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)', '150000', 'FBAR'),
                ('DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)', '180000', 'FBAR'),
                ('DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)', '210000', 'FBAR'),
                ('DIFF(ENS001v3.6.1_d01 DPT FBAR-ENS001v3.6.1_d02 DPT FBAR)', '240000', 'FBAR')]
    assert result == expected


def test_calculate_value_and_ci(settings):
    agg_stat = settings['agg_stat']
    agg_stat.calculate_stats_and_ci()
    result_frame = pd.read_csv(
        agg_stat.params['agg_stat_output'],
        header=[0],
        sep='\t'
    )
    assert result_frame.size == 216
    assert result_frame.shape == (27, 8)
    assert np.allclose(result_frame['stat_value'][2], 192.1304275)
    assert np.allclose(result_frame['stat_value'][20], 0.00362229)
    assert result_frame['stat_btcl'][9] <= result_frame['stat_value'][9] <=result_frame['stat_btcu'][9]
    assert result_frame['stat_btcl'][24] <= result_frame['stat_value'][24] <=result_frame['stat_btcu'][24]


@pytest.fixture
def settings():
    """Initialise values for testing.

    Returns:
        dictionary with values of different type
    """
    params = {'random_seed': 1, 'indy_var': 'fcst_lead',
               'method': 'perc',
              'num_iterations': 100, 'event_equal': 'True',
              'derived_series_1': [
                  ['ENS001v3.6.1_d01 DPT FBAR', 'ENS001v3.6.1_d02 DPT FBAR', 'DIFF']],
              'derived_series_2': [],
              'agg_stat_input': 'data/agg_stat_and_boot_data.data',
              'fcst_var_val_1': {'DPT': ['FBAR']},
              'fcst_var_val_2': {},
              'agg_stat_output': 'data/agg_stat_and_boot_output.data',
              'fixed_vars_vals_input': {'fcst_lev': {'fcst_lev_0': ['P100']}},
              'series_val_1': {'model': ['ENS001v3.6.1_d01', 'ENS001v3.6.1_d02']},
              'series_val_2': {},
              'alpha': 0.05, 'line_type': 'sl1l2',
              'num_threads': -1,
              'indy_vals': ['0', '30000', '60000', '90000',
                            '120000', '150000', '180000', '210000', '240000'],
              'list_stat_1': ['FBAR'],
              'list_stat_2': [],
              'circular_block_bootstrap': False}
    agg_stat = AggStat(params)
    settings_dict = dict()
    settings_dict['agg_stat'] = agg_stat
    return settings_dict
